@@ -15,7 +15,7 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -30,13 +30,6 @@
 #include <sys/stat.h>
 #include <stdarg.h>
 #include <unistd.h>
-#include <openssl/ssl.h>
-#include <openssl/x509.h>
-#include <openssl/err.h>
-
-#ifdef HAVE_NOTIFY
-#include <libnotify/notify.h>
-#endif
 
 #if ENABLE_LDAP
 #include <ldap.h>
@@ -50,50 +43,21 @@
 #include <gtksourceview/gtksource.h>
 #endif
 
-#if HAVE_GCR
-#define GCR_API_SUBJECT_TO_CHANGE
-#include <gcr/gcr.h>
-#endif
-
 #include "misc.h"
 #include "missing.h"
+#include "x509-cert-widget.h"
 #include <glib/gi18n.h>
 
 static GThread *main_thread_id;
 
 
 void
-libbalsa_message(const char *fmt, ...)
+libbalsa_init(void)
 {
-    va_list va_args;
-
-    va_start(va_args, fmt);
-    libbalsa_information_varg(NULL, LIBBALSA_INFORMATION_MESSAGE,
-                              fmt, va_args);
-    va_end(va_args);
-}
-
-void
-libbalsa_init(LibBalsaInformationFunc information_callback)
-{
-#ifdef HAVE_NOTIFY
-    notify_init("Basics");
-#endif
-
     main_thread_id = g_thread_self();
 
-    libbalsa_real_information_func = information_callback;
-
-    g_mime_init(GMIME_ENABLE_RFC2047_WORKAROUNDS);
-
-    GMIME_TYPE_DATA_WRAPPER;
-    GMIME_TYPE_FILTER;
-    GMIME_TYPE_FILTER_CRLF;
-    GMIME_TYPE_PARSER;
-    GMIME_TYPE_STREAM;
-    GMIME_TYPE_STREAM_BUFFER;
-    GMIME_TYPE_STREAM_MEM;
-    GMIME_TYPE_STREAM_NULL;
+    g_mime_init(); /* Registers all GMime types */
+    libbalsa_parser_options_init();
 
     /* Register our types to avoid possible race conditions. See
        output of "valgrind --tool=helgrind --log-file=balsa.log balsa"
@@ -108,7 +72,7 @@ libbalsa_init(LibBalsaInformationFunc information_callback)
     LIBBALSA_TYPE_MESSAGE;
 
     LIBBALSA_TYPE_ADDRESS_BOOK_VCARD;
-    LIBBALSA_TYPE_ADDRESS_BOOK_EXTERN;
+    LIBBALSA_TYPE_ADDRESS_BOOK_EXTERNQ;
     LIBBALSA_TYPE_ADDRESS_BOOK_LDIF;
 #if ENABLE_LDAP
     LIBBALSA_TYPE_ADDRESS_BOOK_LDAP;
@@ -125,32 +89,6 @@ libbalsa_init(LibBalsaInformationFunc information_callback)
 }
 
 
-/* libbalsa_rot:
-   return rot13'ed string.
-*/
-gchar *
-libbalsa_rot(const gchar * pass)
-{
-    gchar *buff;
-    gint len = 0, i = 0;
-
-    /*PKGW: let's do the assert() BEFORE we coredump... */
-
-    len = strlen(pass);
-    buff = g_strdup(pass);
-
-    for (i = 0; i < len; i++) {
-	if ((buff[i] <= 'M' && buff[i] >= 'A')
-	    || (buff[i] <= 'm' && buff[i] >= 'a'))
-	    buff[i] += 13;
-	else if ((buff[i] <= 'Z' && buff[i] >= 'N')
-		 || (buff[i] <= 'z' && buff[i] >= 'n'))
-	    buff[i] -= 13;
-    }
-    return buff;
-}
-
-
 /* libbalsa_guess_email_address:
    Email address can be determined in four ways:
    1. Using the environment variable 'EMAIL'
@@ -305,11 +243,11 @@ static gboolean
 ask_idle(gpointer data)
 {
     AskData* ad = (AskData*)data;
-    printf("ask_idle: ENTER %p\n", data);
+    g_debug("ask_idle: ENTER %p", data);
     ad->res = (ad->cb)(ad->arg);
     ad->done = TRUE;
     g_cond_signal(&ad->condvar);
-    printf("ask_idle: LEAVE %p\n", data);
+    g_debug("ask_idle: LEAVE %p", data);
     return FALSE;
 }
 
@@ -324,11 +262,11 @@ libbalsa_ask(gboolean (*cb)(void *arg), void *arg)
 
     if (!libbalsa_am_i_subthread()) {
         int ret;
-        printf("Main thread asks the following question.\n");
+        g_debug("main thread asks the following question");
         ret = cb(arg);
         return ret;
     }
-    printf("Side thread asks the following question.\n");
+    g_debug("side thread asks the following question");
     g_mutex_init(&ad.lock);
     g_cond_init(&ad.condvar);
     ad.cb  = cb;
@@ -347,145 +285,87 @@ libbalsa_ask(gboolean (*cb)(void *arg), void *arg)
 }
 
 
-static int libbalsa_ask_for_cert_acceptance(X509 *cert,
-					    const char *explanation);
+static int libbalsa_ask_for_cert_acceptance(GTlsCertificate      *cert,
+											GTlsCertificateFlags  errors);
 
-#ifndef HAVE_GCR
-
-static char*
-asn1time_to_string(ASN1_UTCTIME *tm)
-{
-    char buf[64];
-    BIO *bio  = BIO_new(BIO_s_mem());
-    strncpy(buf, _("Invalid date"), sizeof(buf)); buf[sizeof(buf)-1]='\0';
-
-    if(ASN1_TIME_print(bio, tm)) {
-        int cnt;
-        cnt = BIO_read(bio, buf, sizeof(buf)-1);
-        buf[cnt] = '\0';
-    }
-    BIO_free(bio);
-    return g_strdup(buf);
-}
-
-static char*
-x509_get_part (char *line, const char *ndx)
-{
-    static char ret[256];
-    char *c;
-
-    strncpy (ret, _("Unknown"), sizeof (ret)); ret[sizeof(ret)-1]='\0';
-
-    c = strstr(line, ndx);
-    if (c) {
-        char *c2;
-
-        c += strlen (ndx);
-        c2 = strchr (c, '/');
-        if (c2)
-            *c2 = '\0';
-        strncpy (ret, c, sizeof (ret));
-        if (c2)
-            *c2 = '/';
-    }
-
-    return ret;
-}
-static void
-x509_fingerprint (char *s, unsigned len, X509 * cert)
-{
-    unsigned j, i, n, c;
-    unsigned char md[EVP_MAX_MD_SIZE];
-
-
-    X509_digest(cert, EVP_md5(), md, &n);
-    if(len<3*n) n = len/3;
-    for (j=i=0; j<n; j++) {
-        c = (md[j] >>4) & 0xF; s[i++] = c<10 ? c + '0' : c+'A'-10;
-        c = md[j] & 0xF;       s[i++] = c<10 ? c + '0' : c+'A'-10;
-        if(j<n-1) s[i++] = ':';
-    }
-    s[i] = '\0';
-}
-
-#endif  /* HAVE_GCR */
-
-static GList *accepted_certs = NULL; /* certs accepted for this session */
+static GList *accepted_certs = NULL; /* GTlsCertificate items accepted for this session */
 static GMutex certificate_lock;
 
 void
 libbalsa_certs_destroy(void)
 {
 	g_mutex_lock(&certificate_lock);
-    g_list_foreach(accepted_certs, (GFunc)X509_free, NULL);
-    g_list_free(accepted_certs);
+    g_list_free_full(accepted_certs, g_object_unref);
     accepted_certs = NULL;
     g_mutex_unlock(&certificate_lock);
 }
 
-/* compare Example 10-7 in the OpenSSL book */
+
+#define CERT_ACCEPT_NO			0
+#define CERT_ACCEPT_SESSION		1
+#define CERT_ACCEPT_PERMANENT	2
+
+
 gboolean
-libbalsa_is_cert_known(X509* cert, long vfy_result)
+libbalsa_is_cert_known(GTlsCertificate      *cert,
+					   GTlsCertificateFlags  errors)
 {
-    X509 *tmpcert = NULL;
-    FILE *fp;
-    gchar *cert_name;
-    gboolean res = FALSE;
-    GList *lst;
+	gchar *cert_file;
+	GList *cert_db;
+	gboolean cert_ok;
+	GList *lst;
 
-    g_mutex_lock(&certificate_lock);
-    for(lst = accepted_certs; lst; lst = lst->next) {
-        int X509_res = X509_cmp(cert, lst->data);
-        if(X509_res == 0) {
-        	g_mutex_unlock(&certificate_lock);
-            return TRUE;
-	}
-    }
-    
-    cert_name = g_strconcat(g_get_home_dir(), "/.balsa/certificates", NULL);
-
-    fp = fopen(cert_name, "rt");
-    g_free(cert_name);
-    if(fp) {
-        /* 
-        printf("Looking for cert: %s\n", 
-               X509_NAME_oneline(X509_get_subject_name (cert),
-                                 buf, sizeof (buf)));
-        */
-        res = FALSE;
-        while ((tmpcert = PEM_read_X509(fp, NULL, NULL, NULL)) != NULL) {
-            res = X509_cmp(cert, tmpcert)==0;
-            X509_free(tmpcert);
-            if(res) break;
-        }
-        ERR_clear_error();
-        fclose(fp);
-    }
-    g_mutex_unlock(&certificate_lock);
-    
-    if(!res) {
-	const char *reason = X509_verify_cert_error_string(vfy_result);
-	res = libbalsa_ask_for_cert_acceptance(cert, reason);
+	/* check the list of accepted certificates for this session */
 	g_mutex_lock(&certificate_lock);
-	if(res == 2) {
-	    cert_name = g_strconcat(g_get_home_dir(),
-				    "/.balsa/certificates", NULL);
-            libbalsa_assure_balsa_dir();
-	    fp = fopen(cert_name, "a");
-	    if (fp) {
-		if(PEM_write_X509 (fp, cert))
-		    res = TRUE;
-		fclose(fp);
-	    }
-	    g_free(cert_name);
+	for (lst = accepted_certs; lst; lst = lst->next) {
+		if (g_tls_certificate_is_same(cert, G_TLS_CERTIFICATE(lst->data))) {
+			g_mutex_unlock(&certificate_lock);
+			return TRUE;
+		}
 	}
-	if(res == 1)
-	    accepted_certs = 
-		g_list_prepend(accepted_certs, X509_dup(cert));
-	g_mutex_unlock(&certificate_lock);
-    }
 
-    return res;
+	/* check the database of accepted certificates */
+	cert_file = g_build_filename(g_get_home_dir(), ".balsa", "certificates", NULL);
+	cert_db = g_tls_certificate_list_new_from_file(cert_file, NULL);
+	g_mutex_unlock(&certificate_lock);
+
+	cert_ok = FALSE;
+	for (lst = cert_db; !cert_ok && (lst != NULL); lst = g_list_next(lst)) {
+		if (g_tls_certificate_is_same(cert, G_TLS_CERTIFICATE(lst->data))) {
+			cert_ok = TRUE;
+		}
+	}
+	g_list_free_full(cert_db, g_object_unref);
+
+	/* ask the user if the certificate is not in the data base */
+	if (!cert_ok) {
+		int accepted;
+
+		accepted = libbalsa_ask_for_cert_acceptance(cert, errors);
+		g_mutex_lock(&certificate_lock);
+		if (accepted == CERT_ACCEPT_SESSION) {
+			accepted_certs = g_list_prepend(accepted_certs, g_object_ref(cert));
+			cert_ok = TRUE;
+		} else if (accepted == CERT_ACCEPT_PERMANENT) {
+			gchar *pem_data;
+			FILE *fd;
+
+			g_object_get(cert, "certificate-pem", &pem_data, NULL);
+			fd = fopen(cert_file, "a");
+			if (fd != NULL) {
+				fputs(pem_data, fd);
+				fclose(fd);
+			}
+			g_free(pem_data);
+			cert_ok = TRUE;
+		} else {
+			/* nothing to do */
+		}
+		g_mutex_unlock(&certificate_lock);
+	}
+	g_free(cert_file);
+
+	return cert_ok;
 }
 
 /* libbalsa_ask_for_cert_acceptance():
@@ -497,11 +377,10 @@ libbalsa_is_cert_known(X509* cert, long vfy_result)
 
 */
 struct AskCertData {
-    X509 *certificate;
-    const char *explanation;
+    GTlsCertificate *certificate;
+    gchar *explanation;
 };
 
-#if HAVE_GCR
 
 static int
 ask_cert_real(void *data)
@@ -511,11 +390,15 @@ ask_cert_real(void *data)
     GtkWidget *cert_widget;
     GString *str;
     unsigned i;
-    unsigned char *der_buf = NULL;
-    int der_len;
-    GcrCertificate *gcr_cert;
     GtkWidget *label;
 
+    /* never accept if the certificate is broken, resulting in a NULL widget */
+    cert_widget = x509_cert_chain_tls(acd->certificate);
+    if (cert_widget == NULL) {
+    	libbalsa_information(LIBBALSA_INFORMATION_WARNING, _("broken TLS certificate"));
+    	return CERT_ACCEPT_NO;
+    }
+
     dialog = gtk_dialog_new_with_buttons(_("SSL/TLS certificate"),
                                          NULL, /* FIXME: NULL parent */
                                          GTK_DIALOG_MODAL |
@@ -525,11 +408,6 @@ ask_cert_real(void *data)
                                          _("_Reject"), GTK_RESPONSE_CANCEL,
                                          NULL);
     gtk_window_set_role(GTK_WINDOW(dialog), "tls_cert_dialog");
-    der_len = i2d_X509(acd->certificate, &der_buf);
-    gcr_cert = gcr_simple_certificate_new(der_buf, der_len);
-    OPENSSL_free(der_buf);
-    cert_widget = GTK_WIDGET(gcr_certificate_widget_new(gcr_cert));
-    g_object_unref(G_OBJECT(gcr_cert));
 
     str = g_string_new("");
     g_string_printf(str, _("<big><b>Authenticity of this certificate "
@@ -544,166 +422,57 @@ ask_cert_real(void *data)
     gtk_widget_show(label);
     gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(dialog))),
                        cert_widget, TRUE, TRUE, 1);
-    gtk_widget_show(cert_widget);
+    gtk_widget_show_all(cert_widget);
 
     switch(gtk_dialog_run(GTK_DIALOG(dialog))) {
     case 0:
-    	i = 1;
+    	i = CERT_ACCEPT_SESSION;
     	break;
     case 1:
-    	i = 2;
+    	i = CERT_ACCEPT_PERMANENT;
     	break;
     case GTK_RESPONSE_CANCEL:
     default:
-    	i = 0;
+    	i = CERT_ACCEPT_NO;
     	break;
     }
     gtk_widget_destroy(dialog);
+    g_free(acd->explanation);
     return i;
 }
 
-#else
 
 static int
-ask_cert_real(void *data)
-{
-    static const char *part[] =
-        {"/CN=", "/Email=", "/O=", "/OU=", "/L=", "/ST=", "/C="};
-
-    struct AskCertData *acd = (struct AskCertData*)data;
-    X509 *cert = acd->certificate;
-    char buf[256]; /* fingerprint requires EVP_MAX_MD_SIZE*3 */
-    char *name = NULL, *c, *valid_from, *valid_until;
-    GtkWidget* dialog, *label;
-    unsigned i;
-
-    GString* str = g_string_new("");
-
-    g_string_printf(str, _("Authenticity of this certificate "
-                           "could not be verified.\n"
-                           "<b>Reason:</b> %s\n"
-                           "<b>This certificate belongs to:</b>\n"),
-                    acd->explanation);
-
-    name = X509_NAME_oneline(X509_get_subject_name (cert), buf, sizeof (buf));
-    for (i = 0; i < ELEMENTS(part); i++) {
-        g_string_append(str, x509_get_part (name, part[i]));
-        g_string_append_c(str, '\n');
-    }
-
-    g_string_append(str, _("\n<b>This certificate was issued by:</b>\n"));
-    name = X509_NAME_oneline(X509_get_issuer_name(cert), buf, sizeof (buf));
-    for (i = 0; i < ELEMENTS(part); i++) {
-        g_string_append(str, x509_get_part (name, part[i]));
-        g_string_append_c(str, '\n');
-    }
-
-    buf[0] = '\0';
-    x509_fingerprint (buf, sizeof (buf), cert);
-    valid_from  = asn1time_to_string(X509_get_notBefore(cert));
-    valid_until = asn1time_to_string(X509_get_notAfter(cert)),
-    c = g_strdup_printf(_("<b>This certificate is valid</b>\n"
-                          "from %s\n"
-                          "to %s\n"
-                          "<b>Fingerprint:</b> %s"),
-                        valid_from, valid_until,
-                        buf);
-    g_string_append(str, c); g_free(c);
-    g_free(valid_from); g_free(valid_until);
-
-    /* This string uses markup, so we must replace "&" with "&amp;" */
-    c = str->str;
-    while ((c = strchr(c, '&'))) {
-        gssize pos;
-
-        pos = (c - str->str) + 1;
-        g_string_insert(str, pos, "amp;");
-        c = str->str + pos;
-    }
-
-    dialog = gtk_dialog_new_with_buttons(_("SSL/TLS certificate"),
-                                         NULL, /* FIXME: NULL parent */
-                                         GTK_DIALOG_MODAL |
-                                         libbalsa_dialog_flags(),
-                                         _("_Accept Once"), 0,
-                                         _("Accept & _Save"), 1,
-                                         _("_Reject"), GTK_RESPONSE_CANCEL, 
-                                         NULL);
-    gtk_window_set_role(GTK_WINDOW(dialog), "tls_cert_dialog");
-    label = gtk_label_new(str->str);
-    g_string_free(str, TRUE);
-    gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
-    gtk_box_pack_start(GTK_BOX
-                       (gtk_dialog_get_content_area(GTK_DIALOG(dialog))),
-                       label, TRUE, TRUE, 1);
-    gtk_widget_show(label);
-
-    switch(gtk_dialog_run(GTK_DIALOG(dialog))) {
-    case 0: i = 1; break;
-    case 1: i = 2; break;
-    case GTK_RESPONSE_CANCEL:
-    default: i=0; break;
-    }
-    gtk_widget_destroy(dialog);
-    /* Process some events to let the window disappear:
-     * not really necessary but helps with debugging. */
-   while(gtk_events_pending()) 
-        gtk_main_iteration_do(FALSE);
-    printf("%s returns %d\n", __FUNCTION__, i);
-    return i;
-}
-
-#endif	/* HAVE_GCR */
-
-static int
-libbalsa_ask_for_cert_acceptance(X509 *cert, const char *explanation)
+libbalsa_ask_for_cert_acceptance(GTlsCertificate      *cert,
+								 GTlsCertificateFlags  errors)
 {
     struct AskCertData acd;
+    static const gchar *reason_msg[] = {
+		N_("the signing certificate authority is not known"),
+		N_("the certificate does not match the expected identity of the site that it was retrieved from"),
+		N_("the certificate’s activation time is still in the future"),
+		N_("the certificate has expired"),
+		N_("the certificate has been revoked"),
+		N_("the certificate’s algorithm is considered insecure"),
+		N_("an error occurred validating the certificate")
+    };
+    GString *exp_buf = g_string_new(NULL);
+    gsize n;
+
     acd.certificate = cert;
-    acd.explanation = explanation;
-    return libbalsa_ask(ask_cert_real, &acd);
-}
-
-
-static int
-ask_timeout_real(void *data)
-{
-    const char *host = (const char*)data;
-    GtkWidget* dialog;
-    int i;
-
-    dialog = gtk_message_dialog_new(NULL, /* FIXME: NULL parent */
-                                    GTK_DIALOG_MODAL,
-                                    GTK_MESSAGE_INFO,
-                                    GTK_BUTTONS_YES_NO,
-                                    _("Connection to %s timed out. Abort?"),
-                                    host);
-    gtk_window_set_role(GTK_WINDOW(dialog), "timeout_dialog");
-    switch(gtk_dialog_run(GTK_DIALOG(dialog))) {
-    case GTK_RESPONSE_YES: i = 1; break;
-    case GTK_RESPONSE_NO: i = 0; break;
-    default: printf("Unknown response. Defaulting to 'yes'.\n");
-        i = 1;
+    for (n = 0U; n < G_N_ELEMENTS(reason_msg); n++) {
+    	if ((errors & (1U << n)) != 0U) {
+    		g_string_append_printf(exp_buf, "\n\342\200\242 %s", reason_msg[n]);
+    	}
     }
-    gtk_widget_destroy(dialog);
-    /* Process some events to let the window disappear:
-     * not really necessary but helps with debugging. */
-   while(gtk_events_pending()) 
-        gtk_main_iteration_do(FALSE);
-    printf("%s returns %d\n", __FUNCTION__, i);
-    return i;
-}
 
-gboolean
-libbalsa_abort_on_timeout(const char *host)
-{  /* It appears not to be entirely thread safe... Some locks do not
-      get released as they should be. */
-    char *hostname;
-
-    hostname = g_alloca (strlen (host) + 1);
-    strcpy (hostname, host);
-    
-    return libbalsa_ask(ask_timeout_real, hostname) != 0; 
+    if (exp_buf->len > 0U) {
+    	acd.explanation = g_string_free(exp_buf, FALSE);
+    } else {
+    	g_string_free(exp_buf, TRUE);
+    	acd.explanation = g_strdup_printf(_("unknown certificate validation error %u"), (unsigned) errors);
+    }
+    return libbalsa_ask(ask_cert_real, &acd);
 }
 
 
@@ -719,59 +488,6 @@ libbalsa_am_i_subthread(void)
     return g_thread_self() != main_thread_id;
 }
 
-
-#include "libbalsa_private.h"	/* for prototypes */
-static GMutex mailbox_mutex;
-static GCond  mailbox_cond;
-
-/* Lock/unlock a mailbox; no argument checking--we'll assume the caller
- * took care of that. 
- */
-#define LIBBALSA_DEBUG_THREADS FALSE
-void
-libbalsa_lock_mailbox(LibBalsaMailbox * mailbox)
-{
-	GThread *thread_id = g_thread_self();
-
-    g_mutex_lock(&mailbox_mutex);
-
-    if (mailbox->thread_id && mailbox->thread_id != thread_id)
-        while (mailbox->lock)
-            g_cond_wait(&mailbox_cond, &mailbox_mutex);
-
-    /* We'll assume that no-one would destroy a mailbox while we've been
-     * trying to lock it. If they have, we have larger problems than
-     * this reference! */
-    mailbox->lock++;
-    mailbox->thread_id = thread_id;
-
-    g_mutex_unlock(&mailbox_mutex);
-}
-
-void
-libbalsa_unlock_mailbox(LibBalsaMailbox * mailbox)
-{
-	GThread *self;
-
-    self = g_thread_self();
-
-    g_mutex_lock(&mailbox_mutex);
-
-    if (mailbox->lock == 0 || self != mailbox->thread_id) {
-	g_warning("Not holding mailbox lock!!!");
-        g_mutex_unlock(&mailbox_mutex);
-	return;
-    }
-
-    if(--mailbox->lock == 0) {
-        mailbox->thread_id = 0;
-        g_cond_broadcast(&mailbox_cond);
-    }
-
-    g_mutex_unlock(&mailbox_mutex);
-}
-
-
 /* Initialized by the front end. */
 void (*libbalsa_progress_set_text) (LibBalsaProgress * progress,
                                     const gchar * text, guint total);
@@ -1006,11 +722,10 @@ libbalsa_image_error_quark(void)
     return quark;
 }
 
-#if GTK_CHECK_VERSION(3, 12, 0)
-GtkDialogFlags
-libbalsa_dialog_flags(void)
+gboolean
+libbalsa_use_headerbar(void)
 {
-	static GtkDialogFlags dialog_flags = GTK_DIALOG_USE_HEADER_BAR;
+	static gboolean use_headerbar = TRUE;
 	static gint check_done = 0;
 
 	if (g_atomic_int_get(&check_done) == 0) {
@@ -1018,10 +733,15 @@ libbalsa_dialog_flags(void)
 
 		dialog_env = g_getenv("BALSA_DIALOG_HEADERBAR");
 		if ((dialog_env != NULL) && (atoi(dialog_env) == 0)) {
-			dialog_flags = (GtkDialogFlags) 0;
+			use_headerbar = FALSE;
 		}
 		g_atomic_int_set(&check_done, 1);
 	}
-	return dialog_flags;
+	return use_headerbar;
+}
+
+GtkDialogFlags
+libbalsa_dialog_flags(void)
+{
+	return libbalsa_use_headerbar() ? GTK_DIALOG_USE_HEADER_BAR : (GtkDialogFlags) 0;
 }
-#endif
