@@ -15,7 +15,7 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -40,6 +40,11 @@
 #include "mime-stream-shared.h"
 #include <glib/gi18n.h>
 
+#ifdef G_LOG_DOMAIN
+#  undef G_LOG_DOMAIN
+#endif
+#define G_LOG_DOMAIN "mbox-mh"
+
 struct message_info {
     LibBalsaMailboxLocalMessageInfo local_info;
     LibBalsaMessageFlag orig_flags;     /* Has only real flags */
@@ -48,10 +53,6 @@ struct message_info {
 
 #define REAL_FLAGS(flags) (flags & LIBBALSA_MESSAGE_FLAGS_REAL)
 
-static LibBalsaMailboxLocalClass *parent_class = NULL;
-
-static void libbalsa_mailbox_mh_class_init(LibBalsaMailboxMhClass *klass);
-static void libbalsa_mailbox_mh_init(LibBalsaMailboxMh * mailbox);
 static void libbalsa_mailbox_mh_finalize(GObject * object);
 static void libbalsa_mailbox_mh_load_config(LibBalsaMailbox * mailbox,
                                             const gchar * prefix);
@@ -86,33 +87,21 @@ static gboolean libbalsa_mailbox_mh_fetch_message_structure(LibBalsaMailbox
                                                             flags);
 static guint libbalsa_mailbox_mh_total_messages(LibBalsaMailbox * mailbox);
 
+static const guint8 zero[1] = {0};
 
-GType
-libbalsa_mailbox_mh_get_type(void)
-{
-    static GType mailbox_type = 0;
+struct _LibBalsaMailboxMh {
+    LibBalsaMailboxLocal parent;
 
-    if (!mailbox_type) {
-	static const GTypeInfo mailbox_info = {
-	    sizeof(LibBalsaMailboxMhClass),
-            NULL,               /* base_init */
-            NULL,               /* base_finalize */
-	    (GClassInitFunc) libbalsa_mailbox_mh_class_init,
-            NULL,               /* class_finalize */
-            NULL,               /* class_data */
-	    sizeof(LibBalsaMailboxMh),
-            0,                  /* n_preallocs */
-	    (GInstanceInitFunc) libbalsa_mailbox_mh_init
-	};
+    GHashTable* messages_info;
+    GPtrArray* msgno_2_msg_info;
+    gchar* sequences_filename;
+    time_t mtime_sequences;
+    guint last_fileno;
+};
 
-	mailbox_type =
-	    g_type_register_static(LIBBALSA_TYPE_MAILBOX_LOCAL,
-	                           "LibBalsaMailboxMh",
-                                   &mailbox_info, 0);
-    }
-
-    return mailbox_type;
-}
+G_DEFINE_TYPE(LibBalsaMailboxMh,
+              libbalsa_mailbox_mh,
+              LIBBALSA_TYPE_MAILBOX_LOCAL)
 
 static void
 libbalsa_mailbox_mh_class_init(LibBalsaMailboxMhClass * klass)
@@ -125,8 +114,6 @@ libbalsa_mailbox_mh_class_init(LibBalsaMailboxMhClass * klass)
     libbalsa_mailbox_class = LIBBALSA_MAILBOX_CLASS(klass);
     libbalsa_mailbox_local_class = LIBBALSA_MAILBOX_LOCAL_CLASS(klass);
 
-    parent_class = g_type_class_peek_parent(klass);
-
     object_class->finalize = libbalsa_mailbox_mh_finalize;
 
     libbalsa_mailbox_class->load_config = libbalsa_mailbox_mh_load_config;
@@ -178,7 +165,7 @@ lbm_mh_check_files(const gchar * path, gboolean create)
         if (mkdir(path, S_IRWXU)) {
             libbalsa_information(LIBBALSA_INFORMATION_WARNING,
                                  _("Could not create MH directory at %s (%s)"),
-                                 path, strerror(errno));
+                                 path, g_strerror(errno));
             return -1;
         }
 
@@ -190,7 +177,7 @@ lbm_mh_check_files(const gchar * path, gboolean create)
             libbalsa_information
                 (LIBBALSA_INFORMATION_WARNING,
                  _("Could not create MH structure at %s (%s)"),
-                 path, strerror(errno));
+                 path, g_strerror(errno));
             rmdir(path);
             return -1;
         } else
@@ -221,8 +208,6 @@ libbalsa_mailbox_mh_new(const gchar * path, gboolean create)
 
     mailbox = g_object_new(LIBBALSA_TYPE_MAILBOX_MH, NULL);
 
-    mailbox->is_directory = TRUE;
-
     if (libbalsa_mailbox_local_set_path(LIBBALSA_MAILBOX_LOCAL(mailbox),
                                         path, create) != 0) {
 	g_object_unref(mailbox);
@@ -237,7 +222,7 @@ libbalsa_mailbox_mh_finalize(GObject * object)
 {
     LibBalsaMailboxMh *mh = LIBBALSA_MAILBOX_MH(object);
     g_free(mh->sequences_filename);
-    G_OBJECT_CLASS(parent_class)->finalize(object);
+    G_OBJECT_CLASS(libbalsa_mailbox_mh_parent_class)->finalize(object);
 }
 
 static void
@@ -251,7 +236,7 @@ libbalsa_mailbox_mh_load_config(LibBalsaMailbox * mailbox,
     lbm_mh_set_sequences_filename(mh, path);
     g_free(path);
 
-    LIBBALSA_MAILBOX_CLASS(parent_class)->load_config(mailbox, prefix);
+    LIBBALSA_MAILBOX_CLASS(libbalsa_mailbox_mh_parent_class)->load_config(mailbox, prefix);
 }
 
 #define MH_BASENAME(msg_info) \
@@ -286,16 +271,16 @@ lbm_mh_remove_files(LibBalsaMailboxLocal *mailbox)
 
     if (!libbalsa_delete_directory_contents(path)) {
 	libbalsa_information(LIBBALSA_INFORMATION_ERROR,
-			     _("Could not remove contents of %s:\n%s"),
-			     path, strerror(errno));
+			     _("Could not remove contents of %s: %s"),
+			     path, g_strerror(errno));
 	return;
     }
     if ( rmdir(path) == -1 ) {
 	libbalsa_information(LIBBALSA_INFORMATION_ERROR,
-			     _("Could not remove %s:\n%s"),
-			     path, strerror(errno));
+			     _("Could not remove %s: %s"),
+			     path, g_strerror(errno));
     }
-    LIBBALSA_MAILBOX_LOCAL_CLASS(parent_class)->remove_files(mailbox);
+    LIBBALSA_MAILBOX_LOCAL_CLASS(libbalsa_mailbox_mh_parent_class)->remove_files(mailbox);
 }
 
 
@@ -340,7 +325,7 @@ lbm_mh_parse_mailbox(LibBalsaMailboxMh * mh, gboolean add_msg_info)
     GDir *dir;
     const gchar *filename;
 
-    path = libbalsa_mailbox_local_get_path(mh);
+    path = libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mh);
 
     if ((dir = g_dir_open(path, 0, NULL)) == NULL)
 	return;
@@ -399,7 +384,7 @@ lbm_mh_set_flag(LibBalsaMailboxMh * mh, guint fileno, LibBalsaMessageFlag flag)
     msg_info = g_hash_table_lookup(mh->messages_info, GINT_TO_POINTER(fileno));
 
     if (!msg_info) {
-	g_print("MH sequence info for nonexistent message %d\n", fileno);
+	g_debug("MH sequence info for nonexistent message %d", fileno);
 	return;
     }
 
@@ -474,13 +459,11 @@ lbm_mh_parse_sequences(LibBalsaMailboxMh * mailbox)
     g_object_unref(gmime_stream);
     line = g_byte_array_new();
     do {
-	guint8 zero = 0;
-
+	g_byte_array_set_size(line, 0);
 	g_mime_stream_buffer_readln(gmime_stream_buffer, line);
-	g_byte_array_append(line, &zero, 1);
+	g_byte_array_append(line, zero, 1);
 
 	lbm_mh_handle_seq_line(mailbox, (gchar *) line->data);
-	line->len = 0;
     } while (!g_mime_stream_eos(gmime_stream_buffer));
     g_object_unref(gmime_stream_buffer);
     g_byte_array_free(line, TRUE);
@@ -497,14 +480,16 @@ lbm_mh_parse_both(LibBalsaMailboxMh * mh)
 static void
 lbm_mh_free_message_info(struct message_info *msg_info)
 {
-    if (!msg_info)
+    if (msg_info == NULL)
 	return;
-    if (msg_info->local_info.message) {
-	msg_info->local_info.message->mailbox = NULL;
-	msg_info->local_info.message->msgno   = 0;
+
+    if (msg_info->local_info.message != NULL) {
+        libbalsa_message_set_mailbox(msg_info->local_info.message, NULL);
+        libbalsa_message_set_msgno(msg_info->local_info.message, 0);
 	g_object_remove_weak_pointer(G_OBJECT(msg_info->local_info.message),
-				     (gpointer) &msg_info->local_info.message);
+				     (gpointer *) &msg_info->local_info.message);
     }
+
     g_free(msg_info);
 }
 
@@ -515,7 +500,7 @@ libbalsa_mailbox_mh_open(LibBalsaMailbox * mailbox, GError **err)
     struct stat st;
     const gchar* path;
    
-    path = libbalsa_mailbox_local_get_path(mailbox);
+    path = libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mailbox);
 
     if (stat(path, &st) == -1) {
 	g_set_error(err, LIBBALSA_MAILBOX_ERROR, LIBBALSA_MAILBOX_OPEN_ERROR,
@@ -531,14 +516,12 @@ libbalsa_mailbox_mh_open(LibBalsaMailbox * mailbox, GError **err)
     mh->msgno_2_msg_info = g_ptr_array_new();
     mh->last_fileno = 0;
     
-    mailbox->readonly = access (path, W_OK);
-    mailbox->unread_messages = 0;
+    libbalsa_mailbox_set_readonly(mailbox, access(path, W_OK));
+    libbalsa_mailbox_clear_unread_messages(mailbox);
     lbm_mh_parse_both(mh);
 
-#ifdef DEBUG
-    g_print(_("%s: Opening %s Refcount: %d\n"),
-	    "LibBalsaMailboxMh", mailbox->name, mailbox->open_ref);
-#endif
+    g_debug("%s: Opening %s Refcount: %u",
+	    __func__, libbalsa_mailbox_get_name(mailbox), libbalsa_mailbox_get_open_ref(mailbox));
     return TRUE;
 }
 
@@ -559,10 +542,11 @@ lbm_mh_check(LibBalsaMailboxMh * mh, const gchar * path)
 					GMIME_STREAM_BUFFER_BLOCK_READ);
     g_object_unref(gmime_stream);
 
-    line = (GByteArray *) g_array_new(TRUE, FALSE, 1);
+    line = g_byte_array_new();
     do {
-	line->len = 0;
+	g_byte_array_set_size(line, 0);
 	g_mime_stream_buffer_readln(gmime_stream_buffer, line);
+	g_byte_array_append(line, zero, 1);
 
 	if (libbalsa_str_has_prefix((gchar *) line->data,
 				    LibBalsaMailboxMhUnseen)) {
@@ -613,7 +597,8 @@ libbalsa_mailbox_mh_check(LibBalsaMailbox * mailbox)
 {
     struct stat st, st_sequences;
     LibBalsaMailboxMh *mh = LIBBALSA_MAILBOX_MH(mailbox);
-    const gchar *path = libbalsa_mailbox_local_get_path(mailbox);
+    const gchar *path =
+        libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mailbox);
     int modified = 0;
     guint renumber, msgno;
     struct message_info *msg_info;
@@ -689,13 +674,14 @@ libbalsa_mailbox_mh_check(LibBalsaMailbox * mailbox)
     }
     for (msgno = renumber; msgno <= mh->msgno_2_msg_info->len; msgno++) {
 	msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
-	if (msg_info->local_info.message)
-	    msg_info->local_info.message->msgno = msgno;
+	if (msg_info->local_info.message != NULL)
+	    libbalsa_message_set_msgno(msg_info->local_info.message, msgno);
     }
 
-    msgno = mh->msgno_2_msg_info->len;
     lbm_mh_parse_both(mh);
-    libbalsa_mailbox_local_load_messages(mailbox, msgno);
+
+    if (LIBBALSA_MAILBOX_CLASS(libbalsa_mailbox_mh_parent_class)->check != NULL)
+        LIBBALSA_MAILBOX_CLASS(libbalsa_mailbox_mh_parent_class)->check(mailbox);
 }
 
 static void
@@ -713,8 +699,8 @@ libbalsa_mailbox_mh_close_mailbox(LibBalsaMailbox * mailbox,
     g_hash_table_destroy(mh->messages_info);
     mh->messages_info = NULL;
 
-    if (LIBBALSA_MAILBOX_CLASS(parent_class)->close_mailbox)
-        LIBBALSA_MAILBOX_CLASS(parent_class)->close_mailbox(mailbox,
+    if (LIBBALSA_MAILBOX_CLASS(libbalsa_mailbox_mh_parent_class)->close_mailbox)
+        LIBBALSA_MAILBOX_CLASS(libbalsa_mailbox_mh_parent_class)->close_mailbox(mailbox,
                                                             expunge);
 
     /* Now it's safe to free the message info. */
@@ -786,15 +772,11 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     const gchar *path;
     gchar *name_used;
     guint msgno;
-    struct message_info *msg_info;
 
     int fd;
     int sequences_fd;
     GMimeStream *temp_stream;
     const gchar* sequences_filename;
-    GMimeStream *gmime_stream;
-    GMimeStream *gmime_stream_buffer;
-    GByteArray *line;
     gboolean retval = FALSE;
 
     g_return_val_if_fail(LIBBALSA_IS_MAILBOX_MH(mailbox), FALSE);
@@ -819,14 +801,14 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     recent.first  = recent.last = -1;
     recent.line   = g_mime_stream_mem_new();
 
-    path = libbalsa_mailbox_local_get_path(mailbox);
+    path = libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mailbox);
 
     msgno = 1;
     while (msgno <= mh->msgno_2_msg_info->len) {
-	msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
+        struct message_info *msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
 	if (msg_info->local_info.flags == LIBBALSA_MESSAGE_FLAG_INVALID)
 	    msg_info->local_info.flags = msg_info->orig_flags;
-	if (mailbox->state == LB_MAILBOX_STATE_CLOSING)
+	if (libbalsa_mailbox_get_state(mailbox) == LB_MAILBOX_STATE_CLOSING)
 	    msg_info->local_info.flags &= ~LIBBALSA_MESSAGE_FLAG_RECENT;
 
 	if (expunge && (msg_info->local_info.flags
@@ -885,13 +867,13 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 
     /* Renumber */
     for (msgno = 1; msgno <= mh->msgno_2_msg_info->len; msgno++) {
-	msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
-	if (msg_info->local_info.message)
-	    msg_info->local_info.message->msgno = msgno;
+        struct message_info *msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
+	if (msg_info->local_info.message != NULL)
+	    libbalsa_message_set_msgno(msg_info->local_info.message, msgno);
     }
 
     /* open tempfile */
-    path = libbalsa_mailbox_local_get_path(mailbox);
+    path = libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mailbox);
     fd = libbalsa_mailbox_mh_open_temp(path, &name_used);
     if (fd == -1)
     {
@@ -902,15 +884,17 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	g_object_unref(recent.line);
         if (sequences_fd >= 0)
             libbalsa_unlock_file(sequences_filename, sequences_fd, 1);
-#ifdef DEBUG
-        g_print("MH sync “%s”: cannot open temp file.\n", path);
-#endif
+        g_warning("MH sync “%s”: cannot open temp file.", path);
 	return retval;
     }
     temp_stream = g_mime_stream_fs_new(fd);
 
     if (sequences_fd >= 0) {
         /* copy unknown sequences */
+        GMimeStream *gmime_stream;
+        GMimeStream *gmime_stream_buffer;
+        GByteArray *line;
+
         gmime_stream = g_mime_stream_fs_new(sequences_fd);
         gmime_stream_buffer =
             g_mime_stream_buffer_new(gmime_stream,
@@ -920,7 +904,7 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
         do {
             gchar *buf;
 
-            line->len = 0;
+            g_byte_array_set_size(line, 0);
             g_mime_stream_buffer_readln(gmime_stream_buffer, line);
             buf = (gchar *) line->data;
             if (buf &&
@@ -950,9 +934,7 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	g_object_unref(recent.line);
         if (sequences_fd >= 0)
             libbalsa_unlock_file(sequences_filename, sequences_fd, 1);
-#ifdef DEBUG
-        g_print("MH sync “%s”: error finishing sequences line.\n", path);
-#endif
+        g_warning("MH sync “%s”: error finishing sequences line.", path);
 	return retval;
     }
 
@@ -964,10 +946,8 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 
     /* rename tempfile to '.mh_sequences' */
     retval = (libbalsa_safe_rename(name_used, sequences_filename) != -1);
-#ifdef DEBUG
     if (!retval)
-        g_print("MH sync “%s”: error renaming sequences file.\n", path);
-#endif
+        g_warning("MH sync “%s”: error renaming sequences file.", path);
     if (!retval)
 	unlink (name_used);
 
@@ -1002,28 +982,34 @@ libbalsa_mailbox_mh_fetch_message_structure(LibBalsaMailbox * mailbox,
 					    LibBalsaMessage * message,
 					    LibBalsaFetchFlag flags)
 {
-    if (!message->mime_msg) {
+    GMimeMessage *mime_msg;
+
+    mime_msg = libbalsa_message_get_mime_message(message);
+
+    if (mime_msg == NULL) {
 	struct message_info *msg_info;
 	gchar *base_name;
 
 	msg_info =
 	    lbm_mh_message_info_from_msgno(LIBBALSA_MAILBOX_MH(mailbox),
-					   message->msgno);
+					   libbalsa_message_get_msgno(message));
 
 	base_name = MH_BASENAME(msg_info);
-	message->mime_msg =
-	    libbalsa_mailbox_local_get_mime_message(mailbox, base_name, NULL);
+	mime_msg = libbalsa_mailbox_local_get_mime_message(mailbox, base_name, NULL);
 	g_free(base_name);
 
-	if (message->mime_msg) {
-	    g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+        libbalsa_message_set_mime_message(message, mime_msg);
+
+	if (mime_msg != NULL) {
+	    g_mime_object_remove_header(GMIME_OBJECT(mime_msg),
 				        "Status");
-	    g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+	    g_mime_object_remove_header(GMIME_OBJECT(mime_msg),
 				        "X-Status");
+            g_object_unref(mime_msg);
 	}
     }
 
-    return LIBBALSA_MAILBOX_CLASS(parent_class)->
+    return LIBBALSA_MAILBOX_CLASS(libbalsa_mailbox_mh_parent_class)->
         fetch_message_structure(mailbox, message, flags);
 }
 
@@ -1086,8 +1072,7 @@ lbm_mh_add_message(LibBalsaMailboxLocal * local,
     }
     out_stream = g_mime_stream_fs_new(fd);
 
-    crlffilter = g_mime_filter_crlf_new(FALSE,
-                                        FALSE);
+    crlffilter = g_mime_filter_dos2unix_new(FALSE);
     in_stream = g_mime_stream_filter_new(stream);
     g_mime_stream_filter_add(GMIME_STREAM_FILTER(in_stream), crlffilter);
     g_object_unref(crlffilter);
