@@ -15,7 +15,7 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -24,6 +24,7 @@
 
 #include <string.h>
 #include <ctype.h>
+#include <fribidi.h>
 
 #include "libbalsa.h"
 #include "misc.h"
@@ -218,10 +219,10 @@ content2reply(LibBalsaMessageBody * root, gchar * reply_prefix_str,
  * so this doesn't change the content.
  * */
 /* Updated 2003 to implement DelSp=Yes as in
- * http://www.ietf.org/internet-drafts/draft-gellens-format-bis-01.txt
+ * https://www.ietf.org/internet-drafts/draft-gellens-format-bis-01.txt
  */
 /* Now documented in RFC 3676:
- * http://www.ietf.org/rfc/rfc3676.txt
+ * https://www.ietf.org/rfc/rfc3676.txt
  * or
  * http://www.faqs.org/rfcs/rfc3676.html
  */
@@ -596,7 +597,7 @@ get_quote_tag(GtkTextIter * iter)
     for (list = tag_list; list; list = list->next) {
         GtkTextTag *tag = list->data;
         gchar *name;
-        g_object_get(G_OBJECT(tag), "name", &name, NULL);
+        g_object_get(tag, "name", &name, NULL);
         if (name) {
             if (!strncmp(name, "quote-", 6))
                 quote_tag = tag_list->data;
@@ -658,13 +659,8 @@ static gboolean
 is_in_url(GtkTextIter * iter, gint offset, GtkTextTag * url_tag)
 {
     gtk_text_iter_set_line_offset(iter, offset);
-#if GTK_CHECK_VERSION(3, 19, 0)
     return url_tag ? (gtk_text_iter_has_tag(iter, url_tag)
                       && !gtk_text_iter_starts_tag(iter, url_tag)) : FALSE;
-#else                           /* GTK_CHECK_VERSION(3, 20, 0) */
-    return url_tag ? (gtk_text_iter_has_tag(iter, url_tag)
-                      && !gtk_text_iter_begins_tag(iter, url_tag)) : FALSE;
-#endif                          /* GTK_CHECK_VERSION(3, 20, 0) */
 }
 
 /* Remove soft newlines and associated quote strings from num_paras
@@ -857,7 +853,8 @@ get_url_reg(void)
     static struct url_regex_info info = {
         NULL,
         "(((https?|ftps?|nntp)://)|(mailto:|news:))"
-            "(%[0-9A-F]{2}|[-_.!~*';/?:@&=+$,#[:alnum:]])+",
+            "(%[0-9A-F]{2}|[-_.!~*';/?:@&=+$,#[:alnum:]])+"
+            "(%[0-9A-F]{2}|[-_!~*';/?:@&=+$,#[:alnum:]])",
         __func__,
         "url regex compilation failed"
     };
@@ -1213,7 +1210,7 @@ gchar *
 libbalsa_text_to_html(const gchar * title, const gchar * body, const gchar * lang)
 {
     GString * html_body =
-        g_string_new("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
+        g_string_new("<!DOCTYPE HTML>\n");
     gchar * html_subject;
     const gchar * start = body;
     gchar * html_lang;
@@ -1229,12 +1226,12 @@ libbalsa_text_to_html(const gchar * title, const gchar * body, const gchar * lan
         html_lang = g_strdup("x-unknown");
     html_subject = g_markup_escape_text(title, -1);
     g_string_append_printf(html_body, 
-                           "<html lang=\"%s\"><head>\n"
+                           "<html lang=\"%s\">\n<head>\n"
                            "<title>%s</title>\n"
                            "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
-                           "<style type=\"text/css\">\n"
+                           "<style>\n"
                            "  p { margin-top: 0px; margin-bottom: 0px; }\n"
-                           "</style></head>\n"
+                           "</style>\n</head>\n"
                            "<body>\n", html_lang, html_subject);
     g_free(html_subject);
     g_free(html_lang);
@@ -1243,7 +1240,7 @@ libbalsa_text_to_html(const gchar * title, const gchar * body, const gchar * lan
     while (*start) {
         const gchar * eol = strchr(start, '\n');
         const gchar * p = start;
-        PangoDirection direction = PANGO_DIRECTION_NEUTRAL;
+        gboolean is_rtl = FALSE;
         GString * html;
         gsize idx;
 
@@ -1251,8 +1248,19 @@ libbalsa_text_to_html(const gchar * title, const gchar * body, const gchar * lan
             eol = start + strlen(start);
 
         /* find the first real char to determine the paragraph direction */
-        while (p < eol && direction == PANGO_DIRECTION_NEUTRAL) {
-            direction = pango_unichar_direction(g_utf8_get_char(p));
+        /* Use the same logic as fribidi_get_par_direction(), but
+         * without allocating memory for all the gunichars and
+         * FriBidiCharTypes: */
+        while (p < eol) {
+            FriBidiCharType char_type;
+
+            char_type = fribidi_get_bidi_type(g_utf8_get_char(p));
+
+            if (FRIBIDI_IS_LETTER(char_type)) {
+                is_rtl = FRIBIDI_IS_RTL(char_type);
+                break;
+            }
+
             p = g_utf8_next_char(p);
         }
 
@@ -1275,7 +1283,7 @@ libbalsa_text_to_html(const gchar * title, const gchar * body, const gchar * lan
 
         /* append the paragraph, always stating the proper direction */
         g_string_append_printf(html_body, "<p dir=\"%s\">%s</p>\n",
-                               direction == PANGO_DIRECTION_RTL ? "rtl" : "ltr",
+                               is_rtl ? "rtl" : "ltr",
                                *html->str ? html->str : "&nbsp;");
         g_string_free(html, TRUE);
 
@@ -1286,8 +1294,8 @@ libbalsa_text_to_html(const gchar * title, const gchar * body, const gchar * lan
     }
 
     /* close the html context */
-    html_body = g_string_append(html_body, "</body></html>\n");
-    
+    html_body = g_string_append(html_body, "</body>\n</html>\n");
+
     /* return the utf-8 encoded text/html */
     return g_string_free(html_body, FALSE);
 }
