@@ -14,15 +14,13 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
 # include "config.h"
 #endif                          /* HAVE_CONFIG_H */
 #include "rfc3156.h"
 
-#ifdef HAVE_GPGME
-
 #include <string.h>
 #include <gpgme.h>
 
@@ -70,7 +68,7 @@ body_is_type(LibBalsaMessageBody * body, const gchar * type,
 	retval = g_mime_content_type_is_type(content_type, type, sub_type);
     } else {
 	GMimeContentType *content_type =
-	    g_mime_content_type_new_from_string(body->content_type);
+	    g_mime_content_type_parse(libbalsa_parser_options(), body->content_type);
 	retval = g_mime_content_type_is_type(content_type, type, sub_type);
 	g_object_unref(content_type);
     }
@@ -220,7 +218,7 @@ libbalsa_sign_mime_object(GMimeObject ** content, const gchar * rfc822_for,
 	return FALSE;
     }
 
-    g_object_unref(G_OBJECT(*content));
+    g_object_unref(*content);
     *content = GMIME_OBJECT(mps);
     return TRUE;
 }
@@ -274,7 +272,7 @@ libbalsa_encrypt_mime_object(GMimeObject ** content, GList * rfc822_for,
     if (!result) {
 	g_object_unref(encrypted_obj);
     } else {
-    g_object_unref(G_OBJECT(*content));
+    g_object_unref(*content);
     *content = GMIME_OBJECT(encrypted_obj);
     }
     return result;
@@ -308,8 +306,7 @@ libbalsa_sign_encrypt_mime_object(GMimeObject ** content,
 	return FALSE;
 
     /* we want to be able to restore */
-    signed_object = *content;
-    g_object_ref(G_OBJECT(signed_object));
+    signed_object = g_object_ref(*content);
 
     if (!libbalsa_sign_mime_object(&signed_object, rfc822_signer, protocol,
 				   parent, error))
@@ -317,10 +314,10 @@ libbalsa_sign_encrypt_mime_object(GMimeObject ** content,
 
     if (!libbalsa_encrypt_mime_object(&signed_object, rfc822_for, protocol,
 				      always_trust, parent, error)) {
-	g_object_unref(G_OBJECT(signed_object));
+	g_object_unref(signed_object);
 	return FALSE;
     }
-    g_object_unref(G_OBJECT(*content));
+    g_object_unref(*content);
     *content = signed_object;
 
     return TRUE;
@@ -340,6 +337,7 @@ gboolean
 libbalsa_body_check_signature(LibBalsaMessageBody * body,
 			      gpgme_protocol_t protocol)
 {
+    LibBalsaMailbox *mailbox;
     GError *error = NULL;
     GMimeGpgmeSigstat *result;
 
@@ -358,10 +356,11 @@ libbalsa_body_check_signature(LibBalsaMessageBody * body,
             (GMIME_MULTIPART(body->mime_part)) < 2))
         return FALSE;
     if (body->parts->next->sig_info)
-	g_object_unref(G_OBJECT(body->parts->next->sig_info));
+	g_object_unref(body->parts->next->sig_info);
 
     /* verify the signature */
-    libbalsa_mailbox_lock_store(body->message->mailbox);
+    mailbox = libbalsa_message_get_mailbox(body->message);
+    libbalsa_mailbox_lock_store(mailbox);
     result = g_mime_gpgme_mps_verify(GMIME_MULTIPART_SIGNED(body->mime_part), &error);
     if (!result) {
 	if (error) {
@@ -375,7 +374,7 @@ libbalsa_body_check_signature(LibBalsaMessageBody * body,
     }
 
     body->parts->next->sig_info = result;
-    libbalsa_mailbox_unlock_store(body->message->mailbox);
+    libbalsa_mailbox_unlock_store(mailbox);
     return TRUE;
 }
 
@@ -388,6 +387,7 @@ libbalsa_body_check_signature(LibBalsaMessageBody * body,
 LibBalsaMessageBody *
 libbalsa_body_decrypt(LibBalsaMessageBody *body, gpgme_protocol_t protocol, GtkWindow *parent)
 {
+    LibBalsaMailbox *mailbox;
     GMimeObject *mime_obj = NULL;
     GError *error = NULL;
     LibBalsaMessage *message;
@@ -420,7 +420,8 @@ libbalsa_body_decrypt(LibBalsaMessageBody *body, gpgme_protocol_t protocol, GtkW
     		smime_encrypted = body->was_encrypted;
     }
 
-    libbalsa_mailbox_lock_store(body->message->mailbox);
+    mailbox = libbalsa_message_get_mailbox(body->message);
+    libbalsa_mailbox_lock_store(mailbox);
     if (protocol == GPGME_PROTOCOL_OpenPGP) {
     	mime_obj =
     		g_mime_gpgme_mpe_decrypt(GMIME_MULTIPART_ENCRYPTED(body->mime_part),
@@ -430,7 +431,7 @@ libbalsa_body_decrypt(LibBalsaMessageBody *body, gpgme_protocol_t protocol, GtkW
     		g_mime_application_pkcs7_decrypt_verify(GMIME_PART(body->mime_part),
     			&sig_state, parent, &error);
     }
-    libbalsa_mailbox_unlock_store(body->message->mailbox);
+    libbalsa_mailbox_unlock_store(mailbox);
 
     /* check the result */
     if (mime_obj == NULL) {
@@ -456,14 +457,14 @@ libbalsa_body_decrypt(LibBalsaMessageBody *body, gpgme_protocol_t protocol, GtkW
     	body->was_encrypted = smime_encrypted;
     }
     if (body->was_encrypted)
-        body->message->prot_state = LIBBALSA_MSG_PROTECT_CRYPT;
+        libbalsa_message_set_protect_state(body->message, LIBBALSA_MSG_PROTECT_CRYPT);
 
     libbalsa_message_body_set_mime_body(body, mime_obj);
     if (sig_state) {
-	if (sig_state->status != GPG_ERR_NOT_SIGNED)
+	if (g_mime_gpgme_sigstat_status(sig_state) != GPG_ERR_NOT_SIGNED)
 	    body->sig_info = sig_state;
 	else
-	    g_object_unref(G_OBJECT(sig_state));
+	    g_object_unref(sig_state);
     }
 
     return body;
@@ -547,14 +548,14 @@ libbalsa_rfc2440_verify(GMimePart * part, GMimeGpgmeSigstat ** sig_info)
 	    retval = GPG_ERR_GENERAL;
 	}
     } else
-	retval = result->status;
+	retval = g_mime_gpgme_sigstat_status(result);
 
     /* return the signature info if requested */
     if (result) {
 	if (sig_info)
 	    *sig_info = result;
 	else
-	    g_object_unref(G_OBJECT(result));
+	    g_object_unref(result);
     }
     return retval;
 }
@@ -603,16 +604,16 @@ libbalsa_rfc2440_decrypt(GMimePart * part, GMimeGpgmeSigstat ** sig_info,
 	    retval = GPG_ERR_GENERAL;
 	}
     } else {
-	if (result->status == GPG_ERR_NOT_SIGNED)
+	if (g_mime_gpgme_sigstat_status(result) == GPG_ERR_NOT_SIGNED)
 	    retval = GPG_ERR_NO_ERROR;
 	else
-	    retval = result->status;
+	    retval = g_mime_gpgme_sigstat_status(result);
 
 	/* return the signature info if requested */
-	if (sig_info && result->status != GPG_ERR_NOT_SIGNED)
+	if (sig_info && g_mime_gpgme_sigstat_status(result) != GPG_ERR_NOT_SIGNED)
 	    *sig_info = result;
 	else
-	    g_object_unref(G_OBJECT(result));
+	    g_object_unref(result);
     }
 
     return retval;
@@ -620,42 +621,41 @@ libbalsa_rfc2440_decrypt(GMimePart * part, GMimeGpgmeSigstat ** sig_info,
 
 
 /* conversion of status values to human-readable messages */
-const gchar *
+gchar *
 libbalsa_gpgme_sig_stat_to_gchar(gpgme_error_t stat)
 {
-    switch (stat) {
-    case GPG_ERR_NO_ERROR:
-	return _("The signature is valid.");
-    case GPG_ERR_SIG_EXPIRED:
-	return _("The signature is valid but expired.");
-    case GPG_ERR_KEY_EXPIRED:
-	return _
-	    ("The signature is valid but the key used to verify the signature has expired.");
-    case GPG_ERR_CERT_REVOKED:
-	return _
-	    ("The signature is valid but the key used to verify the signature has been revoked.");
-    case GPG_ERR_BAD_SIGNATURE:
-	return _
-	    ("The signature is invalid.");
-    case GPG_ERR_NO_PUBKEY:
-	return
-	    _("The signature could not be verified due to a missing key.");
-    case GPG_ERR_NO_DATA:
-	return _("This part is not a real PGP signature.");
-    case GPG_ERR_INV_ENGINE:
-	return _
-	    ("The signature could not be verified due to an invalid crypto engine.");
-    case GPG_ERR_TRY_AGAIN:
-	return _
-	    ("GnuPG is rebuilding the trust database and is currently unavailable.");
-    default: {
-	gchar errbuf[4096];		/* should be large enough... */
+	switch (stat) {
+	case GPG_ERR_NO_ERROR:
+		return g_strdup(_("The signature is valid."));
+	case GPG_ERR_SIG_EXPIRED:
+		return g_strdup(_("The signature is valid but expired."));
+	case GPG_ERR_KEY_EXPIRED:
+		return g_strdup(_("The signature is valid but the key used to verify the signature has expired."));
+	case GPG_ERR_CERT_REVOKED:
+		return g_strdup(_("The signature is valid but the key used to verify the signature has been revoked."));
+	case GPG_ERR_BAD_SIGNATURE:
+		return g_strdup(_("The signature is invalid."));
+	case GPG_ERR_NO_PUBKEY:
+		return g_strdup(_("The signature could not be verified due to a missing key."));
+	case GPG_ERR_NO_DATA:
+		return g_strdup(_("This part is not a real signature."));
+	case GPG_ERR_INV_ENGINE:
+		return g_strdup(_("The signature could not be verified due to an invalid crypto engine."));
+	case GPG_ERR_TRY_AGAIN:
+		return g_strdup(_("GnuPG is rebuilding the trust database and is currently unavailable."));
+	case GPG_ERR_MULT_SIGNATURES:
+		return g_strdup(_("The signature contains multiple signers, this may be a forgery."));
+	default: {
+		gchar errbuf[4096];		/* should be large enough... */
 
-	gpgme_strerror_r(stat, errbuf, sizeof(errbuf));
-	g_debug("%s: stat %d: %s %s", __func__, stat, gpgme_strsource(stat), errbuf);
-	return _("An error prevented the signature verification.");
-    }
-    }
+		gpgme_strerror_r(stat, errbuf, sizeof(errbuf));
+		if (gpgme_err_source(stat) != GPG_ERR_SOURCE_UNKNOWN) {
+			return g_strdup_printf(_("An error prevented the signature verification: %s: %s"), gpgme_strsource(stat), errbuf);
+		} else {
+			return g_strdup_printf(_("An error prevented the signature verification: %s"), errbuf);
+		}
+	}
+	}
 }
 
 
@@ -753,7 +753,7 @@ have_pub_key_for(gpgme_ctx_t          gpgme_ctx,
     		result = FALSE;
     		/* enclose the mail address into "<...>" to perform an exact search */
     		mail_name = g_strconcat("<", INTERNET_ADDRESS_MAILBOX(ia)->addr, ">", NULL);
-    		if (libbalsa_gpgme_list_keys(gpgme_ctx, &keys, NULL, mail_name, FALSE, FALSE, NULL)) {
+    		if (libbalsa_gpgme_list_keys(gpgme_ctx, &keys, NULL, mail_name, FALSE, FALSE, FALSE, NULL)) {
     			if (keys != NULL) {
     				result = TRUE;
     				g_list_free_full(keys, (GDestroyNotify) gpgme_key_unref);
@@ -765,5 +765,3 @@ have_pub_key_for(gpgme_ctx_t          gpgme_ctx,
 
     return result;
 }
-
-#endif				/* HAVE_GPGME */
