@@ -14,7 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -68,12 +68,14 @@ struct ABMainWindow {
     LibBalsaAddress *displayed_address;
 
     GMenu *file_menu;
+    GMenu *books_menu;
 } contacts_app;
 
 
 static void
 bab_cleanup(void)
 {
+    g_object_unref(contacts_app.books_menu);
     gtk_main_quit();
 }
 
@@ -122,35 +124,40 @@ bab_load_cb(LibBalsaAddressBook *libbalsa_ab,
             LibBalsaAddress *address, GtkTreeModel *model)
 {
     GtkTreeIter iter;
+    guint n_addrs;
 
     g_return_if_fail(LIBBALSA_IS_ADDRESS_BOOK(libbalsa_ab));
 
     if (address == NULL)
 	return;
 
-    if (libbalsa_address_is_dist_list(libbalsa_ab, address)) {
+    n_addrs = libbalsa_address_get_n_addrs(address);
+    if (libbalsa_address_book_get_dist_list_mode(libbalsa_ab)
+        && n_addrs > 1) {
         gchar *address_string = libbalsa_address_to_gchar(address, -1);
 
         gtk_list_store_prepend(GTK_LIST_STORE(model), &iter);
         /* GtkListStore refs address, and unrefs it when cleared  */
         gtk_list_store_set(GTK_LIST_STORE(model), &iter,
-                           LIST_COLUMN_NAME, address->full_name,
+                           LIST_COLUMN_NAME, libbalsa_address_get_full_name(address),
                            LIST_COLUMN_ADDRSPEC, address_string,
                            LIST_COLUMN_ADDRESS, address,
                            -1);
 
 	g_free(address_string);
     } else {
-        GList *l;
+        guint n;
+
+	for (n = 0; n < n_addrs; ++n) {
+            const gchar *addr = libbalsa_address_get_nth_addr(address, n);
 
-	for (l = address->address_list; l; l = l->next) {
             gtk_list_store_prepend(GTK_LIST_STORE(model), &iter);
             /* GtkListStore refs address once for each address in
              * the list, and unrefs it the same number of times when
              * cleared */
             gtk_list_store_set(GTK_LIST_STORE(model), &iter,
-                               LIST_COLUMN_NAME, address->full_name,
-                               LIST_COLUMN_ADDRSPEC, l->data,
+                               LIST_COLUMN_NAME, libbalsa_address_get_full_name(address),
+                               LIST_COLUMN_ADDRSPEC, addr,
                                LIST_COLUMN_ADDRESS, address,
                                -1);
 	}
@@ -175,7 +182,7 @@ bab_set_address_book(LibBalsaAddressBook * ab,
          libbalsa_address_book_load(ab, filter,
                                     (LibBalsaAddressBookLoadFunc)
                                     bab_load_cb, model)) != LBABERR_OK) {
-        printf("error loading address book from %s: %d\n", ab->name,
+        g_warning("error loading address book from %s: %d", libbalsa_address_book_get_name(ab),
                ab_err);
     }
 
@@ -190,7 +197,7 @@ bab_window_set_title(LibBalsaAddressBook * address_book)
 
     if (LIBBALSA_IS_ADDRESS_BOOK_VCARD(address_book))
         type = "vCard";
-    else if (LIBBALSA_IS_ADDRESS_BOOK_EXTERN(address_book))
+    else if (LIBBALSA_IS_ADDRESS_BOOK_EXTERNQ(address_book))
         type = "External query";
     else if (LIBBALSA_IS_ADDRESS_BOOK_LDIF(address_book))
         type = "LDIF";
@@ -208,7 +215,7 @@ bab_window_set_title(LibBalsaAddressBook * address_book)
 #endif /* HAVE_RUBRICA */
 
     title =
-        g_strconcat(type, _(" address book: "), address_book->name, NULL);
+        g_strconcat(type, _(" address book: "), libbalsa_address_book_get_name(address_book), NULL);
     gtk_window_set_title(contacts_app.window, title);
     g_free(title);
 }
@@ -225,7 +232,7 @@ address_book_change_state(GSimpleAction * action,
     value = g_variant_get_string(state, NULL);
     for (l = contacts_app.address_book_list; l; l = l->next) {
         address_book = l->data;
-        if (address_book && strcmp(value, address_book->name) == 0)
+        if (address_book && strcmp(value, libbalsa_address_book_get_name(address_book)) == 0)
             break;
     }
 
@@ -251,63 +258,47 @@ address_changed_cb(struct ABMainWindow *aw)
 static void
 set_address_book_menu_items(void)
 {
-    GString *string;
     GList *l;
     guint pos;
-    gchar *s;
-    GtkBuilder *builder;
-    GMenuModel *menu_model;
+    GMenu *menu = contacts_app.books_menu;
 
-    pos = g_menu_model_get_n_items(G_MENU_MODEL (contacts_app.file_menu));
-    g_menu_remove(contacts_app.file_menu, --pos);
+    if (menu == NULL) {
+        contacts_app.books_menu = menu = g_menu_new();
+        g_menu_append_section(contacts_app.file_menu, NULL, G_MENU_MODEL(menu));
+    } else {
+        g_menu_remove_all(menu);
+    }
 
     pos = 0;
-    string = g_string_new(NULL);
-    g_string_append(string, "<interface>");
-    g_string_append(string, "<menu id='address-book-menu'>");
-    g_string_append(string, "<section>");
-    for (l = contacts_app.address_book_list; l; l = l->next) {
+    for (l = contacts_app.address_book_list; l != NULL; l = l->next) {
         LibBalsaAddressBook *address_book = l->data;
+        const gchar *name;
+        gchar *label;
+        gchar *detailed_action;
+        gchar *accel;
+        GMenuItem *item;
 
-        if (!address_book)
+        if (address_book == NULL)
             continue;
 
-        g_string_append(string, "<item>");
+        name = libbalsa_address_book_get_name(address_book);
 
-        g_string_append(string, "<attribute name='label'>");
-        g_string_append_printf(string, "_%d:%s", ++pos,
-                               address_book->name);
-        g_string_append(string, "</attribute>");
+        label = g_strdup_printf("_%d:%s", ++pos, name);
+        detailed_action = g_strdup_printf("win.address-book::%s", name);
+        item = g_menu_item_new(label, detailed_action);
+        g_free(detailed_action);
+        g_free(label);
 
-        g_string_append(string, "<attribute name='action'>");
-        g_string_append(string, "win.address-book");
-        g_string_append(string, "</attribute>");
+        accel = g_strdup_printf("<Primary>%d", pos);
+        g_menu_item_set_attribute(item, "accel", "s", accel);
+        g_free(accel);
 
-        g_string_append(string, "<attribute name='target'>");
-        g_string_append(string, address_book->name);
-        g_string_append(string, "</attribute>");
-
-        g_string_append(string, "<attribute name='accel'>");
-        g_string_append_printf(string, "&lt;Primary&gt;%d", pos);
-        g_string_append(string, "</attribute>");
-
-        g_string_append(string, "</item>");
+        g_menu_append_item(menu, item);
+        g_object_unref(item);
     }
-    g_string_append(string, "</section>");
-    g_string_append(string, "</menu>");
-    g_string_append(string, "</interface>");
-    s = g_string_free(string, FALSE);
 
-    builder = gtk_builder_new_from_string(s, -1);
-    g_free(s);
-
-    menu_model =
-        G_MENU_MODEL(gtk_builder_get_object(builder, "address-book-menu"));
-    g_menu_append_section(contacts_app.file_menu, NULL, menu_model);
     libbalsa_window_set_accels(GTK_APPLICATION_WINDOW(contacts_app.window),
-                               menu_model);
-
-    g_object_unref(builder);
+                               G_MENU_MODEL(menu));
 }
 
 static gboolean
@@ -337,17 +328,20 @@ get_unused_group(const gchar * prefix)
 static void
 address_book_change(LibBalsaAddressBook * address_book, gboolean append)
 {
+    const gchar *config_prefix;
     gchar *group;
 
-    if (append)
+    if (append) {
         contacts_app.address_book_list =
             g_list_append(contacts_app.address_book_list, address_book);
+    }
 
     set_address_book_menu_items();
     bab_window_set_title(address_book);
 
-    group = address_book->config_prefix ?
-        g_strdup(address_book->config_prefix) :
+    config_prefix = libbalsa_address_book_get_config_prefix(address_book);
+    group = config_prefix != NULL ?
+        g_strdup(config_prefix) :
         get_unused_group(ADDRESS_BOOK_SECTION_PREFIX);
     libbalsa_address_book_save_config(address_book, group);
     g_free(group);
@@ -366,12 +360,12 @@ file_new_vcard_activated(GSimpleAction * action,
 }
 
 static void
-file_new_extern_activated(GSimpleAction * action,
+file_new_externq_activated(GSimpleAction * action,
                           GVariant      * state,
                           gpointer        user_data)
 {
     balsa_address_book_config_new_from_type
-        (LIBBALSA_TYPE_ADDRESS_BOOK_EXTERN, address_book_change,
+        (LIBBALSA_TYPE_ADDRESS_BOOK_EXTERNQ, address_book_change,
          contacts_app.window);
 }
 
@@ -441,15 +435,17 @@ file_delete_activated(GSimpleAction * action,
                       gpointer        user_data)
 {
     LibBalsaAddressBook *address_book;
+    const gchar *config_prefix;
     GList *list;
 
-    if (!(address_book = contacts_app.address_book)
-        || !g_list_next(contacts_app.address_book_list))
+    if ((address_book = contacts_app.address_book) == NULL
+        || contacts_app.address_book_list->next == NULL)
         return;
 
-    if (address_book->config_prefix) {
-        libbalsa_conf_remove_group(address_book->config_prefix);
-        libbalsa_conf_private_remove_group(address_book->config_prefix);
+    config_prefix = libbalsa_address_book_get_config_prefix(address_book);
+    if (config_prefix != NULL) {
+        libbalsa_conf_remove_group(config_prefix);
+        libbalsa_conf_private_remove_group(config_prefix);
         libbalsa_conf_queue_sync();
     }
 
@@ -466,7 +462,7 @@ file_delete_activated(GSimpleAction * action,
     if (!list)
         return;
 
-    contacts_app.address_book = list->data;
+    contacts_app.address_book = address_book;
     set_address_book_menu_items();
 }
 
@@ -558,7 +554,7 @@ get_main_menu(GtkApplication * application)
 {
     static GActionEntry win_entries[] = {
         {"file-new-vcard",      file_new_vcard_activated},
-        {"file-new-external",   file_new_extern_activated},
+        {"file-new-external",   file_new_externq_activated},
         {"file-new-ldif",       file_new_ldif_activated},
 #if ENABLE_LDAP
         {"file-new-ldap",       file_new_ldap_activated},
@@ -575,8 +571,7 @@ get_main_menu(GtkApplication * application)
         {"entry-new",           entry_new_activated},
         {"entry-delete",        entry_delete_activated},
         {"help-about",          help_about_activated},
-        {"address-book",        libbalsa_radio_activated, "s", "''",
-            address_book_change_state},
+        {"address-book",        NULL, "s", "''", address_book_change_state},
     };
     GtkBuilder *builder;
     const gchar resource_path[] = "/org/desktop/BalsaAb/ab-main.ui";
@@ -590,7 +585,7 @@ get_main_menu(GtkApplication * application)
         contacts_app.file_menu =
             G_MENU(gtk_builder_get_object(builder, "file-menu"));
     } else {
-        g_print("%s error: %s\n", __func__, err->message);
+        g_critical("%s error: %s", __func__, err->message);
         g_error_free(err);
     }
     g_object_unref(builder);
@@ -662,7 +657,7 @@ list_row_activated_cb(GtkTreeView *tree, gpointer data)
     address = LIBBALSA_ADDRESS(g_value_get_object(&gv));
     if (address) {
         ab_set_edit_widget(address, TRUE);
-	printf("Switch page..\n");
+	g_debug("Switch page");
 	gtk_notebook_set_current_page(GTK_NOTEBOOK(contacts_app.notebook), 1);
     } else
         ab_clear_edit_widget();
@@ -698,9 +693,9 @@ addrlist_drag_get_cb(GtkWidget* widget, GdkDragContext* drag_context,
                                sizeof(LibBalsaAddress*));
         break;
     case LIBBALSA_ADDRESS_TRG_STRING:
-        g_print("Text/plain cannot be sent.\n");
+        g_warning("Text/plain cannot be sent.");
         break;
-    default: g_print("Do not know what to do!\n");
+    default: g_debug("Do not know what to do!");
     }
 }
 
@@ -730,9 +725,9 @@ bab_window_list_new(void)
     g_object_unref(store);
     selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
     gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
-    g_signal_connect(G_OBJECT(selection), "changed",
+    g_signal_connect(selection, "changed",
                      G_CALLBACK(list_selection_changed_cb), NULL);
-    g_signal_connect(G_OBJECT(tree), "row-activated",
+    g_signal_connect(tree, "row-activated",
                      G_CALLBACK(list_row_activated_cb), NULL);
 
     renderer = gtk_cell_renderer_text_new();
@@ -748,7 +743,7 @@ bab_window_list_new(void)
                         GDK_BUTTON1_MASK,
                         libbalsa_address_target_list, 2,
                         GDK_ACTION_COPY);
-    g_signal_connect(G_OBJECT(tree), "drag-data-get",
+    g_signal_connect(tree, "drag-data-get",
                      G_CALLBACK(addrlist_drag_get_cb), NULL);
 
     renderer = gtk_cell_renderer_text_new();
@@ -842,17 +837,17 @@ bab_get_edit_button_box(struct ABMainWindow *abmw)
     gtk_container_add(GTK_CONTAINER(box),
                       abmw->apply_button =
                       gtk_button_new_with_mnemonic(_("_Apply")));
-    g_signal_connect(G_OBJECT(abmw->apply_button), "clicked",
+    g_signal_connect(abmw->apply_button, "clicked",
                      G_CALLBACK(apply_button_cb), (gpointer) NULL);
     gtk_container_add(GTK_CONTAINER(box),
                       abmw->remove_button =
                       gtk_button_new_with_mnemonic(_("_Remove")));
-    g_signal_connect(G_OBJECT(abmw->remove_button), "clicked",
+    g_signal_connect(abmw->remove_button, "clicked",
                      G_CALLBACK(remove_button_cb), (gpointer) NULL);
     gtk_container_add(GTK_CONTAINER(box),
                       abmw->cancel_button =
                       gtk_button_new_with_mnemonic(_("_Cancel")));
-    g_signal_connect(G_OBJECT(abmw->cancel_button), "clicked",
+    g_signal_connect(abmw->cancel_button, "clicked",
                      G_CALLBACK(cancel_button_cb), abmw);
     return box;
 }
@@ -893,23 +888,32 @@ bab_get_filter_box(void)
                                                    GTK_ICON_SIZE_BUTTON));
     gtk_box_pack_start(GTK_BOX(search_hbox), button, FALSE, FALSE, 1);
 
-    g_signal_connect(G_OBJECT(find_entry), "activate",
+    g_signal_connect(find_entry, "activate",
                      G_CALLBACK(bab_filter_entry_activate),
                      button);
-    g_signal_connect_swapped(G_OBJECT(button), "clicked",
+    g_signal_connect_swapped(button, "clicked",
                              G_CALLBACK(bab_filter_entry_activate),
                              find_entry);
-    g_signal_connect(G_OBJECT(find_entry), "changed",
+    g_signal_connect(find_entry, "changed",
                              G_CALLBACK(bab_filter_entry_changed),
                              button);
     return search_hbox;
 }
+
 static gboolean
-ew_key_pressed(GtkEntry * entry, GdkEventKey * event, struct ABMainWindow *abmw)
+ew_key_pressed(GtkEventControllerKey *controller,
+               guint                  keyval,
+               guint                  keycode,
+               GdkModifierType        state,
+               gpointer               user_data)
 {
-    if (event->keyval != GDK_KEY_Escape)
+    struct ABMainWindow *abmw = user_data;
+
+    if (keyval != GDK_KEY_Escape)
 	return FALSE;
+
     gtk_button_clicked(GTK_BUTTON(abmw->cancel_button));
+
     return TRUE;
 }
 
@@ -921,6 +925,7 @@ bab_window_new(GtkApplication * application)
     GtkWidget *scroll;
     GtkWidget *browse_widget;
     GtkWidget *edit_widget;
+    GtkEventController *key_controller;
 
     contacts_app.window =
         GTK_WINDOW(wnd = gtk_application_window_new(application));
@@ -969,22 +974,19 @@ bab_window_new(GtkApplication * application)
 			     gtk_label_new(_("Edit")));
 
     /*
-    g_signal_connect(G_OBJECT(find_entry), "changed",
+    g_signal_connect(find_entry, "changed",
 		     G_CALLBACK(balsa_ab_window_find), ab);
     */
-    g_signal_connect(wnd, "key-press-event",
+    key_controller = gtk_event_controller_key_new(wnd);
+    g_signal_connect(key_controller, "key-pressed",
 		     G_CALLBACK(ew_key_pressed), &contacts_app);
+
     gtk_window_set_default_size(GTK_WINDOW(wnd), 500, 400);
 
     gtk_widget_show_all(wnd);
     return wnd;
 }
 
-static gboolean
-bab_delete_ok(void)
-{
-    return FALSE;
-}
 /* -------------------------- main --------------------------------- */
 static void
 ab_warning(const char *fmt, ...)
@@ -1007,7 +1009,7 @@ static void
 bab_init(void)
 {
     LIBBALSA_TYPE_ADDRESS_BOOK_VCARD;
-    LIBBALSA_TYPE_ADDRESS_BOOK_EXTERN;
+    LIBBALSA_TYPE_ADDRESS_BOOK_EXTERNQ;
     LIBBALSA_TYPE_ADDRESS_BOOK_LDIF;
 #if ENABLE_LDAP
     LIBBALSA_TYPE_ADDRESS_BOOK_LDAP;
@@ -1021,24 +1023,22 @@ bab_init(void)
     memset(&contacts_app, 0, sizeof(contacts_app));
 }
 
-static void
-information_real(void)
-{
-    /* FIXME */
-}
-
 static void
 bab_set_intial_address_book(LibBalsaAddressBook * ab,
                             GtkWidget           * window)
 {
-    GAction *action;
+    if (ab != NULL) {
+        GAction *action;
 
-    action =
-        g_action_map_lookup_action(G_ACTION_MAP(window), "address-book");
-    g_action_change_state(action, g_variant_new_string(ab->name));
+        action =
+            g_action_map_lookup_action(G_ACTION_MAP(window),
+                                       "address-book");
+        g_action_change_state(action,
+                              g_variant_new_string
+                              (libbalsa_address_book_get_name(ab)));
+    }
 }
 
-#if GTK_CHECK_VERSION(3, 12, 0)
 GtkDialogFlags
 libbalsa_dialog_flags(void)
 {
@@ -1056,7 +1056,44 @@ libbalsa_dialog_flags(void)
 	}
 	return dialog_flags;
 }
-#endif
+
+/*
+ * Set up GNotification for libbalsa
+ */
+
+#define BALSA_AB_NOTIFICATION "balsa-ab-notification"
+
+static void
+balsa_ab_notification_notify_cb(GNotification *notification, GApplication *application)
+{
+    gboolean send;
+
+    send = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(notification), "send"));
+    if (send) {
+        g_application_send_notification(application,
+                                        BALSA_AB_NOTIFICATION, notification);
+    } else {
+        g_application_withdraw_notification(application, BALSA_AB_NOTIFICATION);
+    }
+}
+
+static void
+balsa_ab_notification_shutdown_cb(GApplication *application, GNotification *notification)
+{
+    g_object_unref(notification);
+}
+
+static void
+balsa_ab_setup_libbalsa_notification(GApplication *application)
+{
+    GNotification *notification;
+
+    notification = libbalsa_notification_new("BalsaAb");
+    g_signal_connect(notification, "notify",
+                     G_CALLBACK(balsa_ab_notification_notify_cb), application);
+    g_signal_connect(application, "shutdown",
+                     G_CALLBACK(balsa_ab_notification_shutdown_cb), notification);
+}
 
 int
 main(int argc, char *argv[])
@@ -1065,11 +1102,14 @@ main(int argc, char *argv[])
     LibBalsaAddressBook *ab;
     GtkWidget *ab_window;
     GList *l;
+    GError *error = NULL;
 
     application =
         gtk_application_new("org.desktop.BalsaAb", G_APPLICATION_FLAGS_NONE);
-    if (!g_application_register(G_APPLICATION(application), NULL, NULL))
-        g_message("Could not register address book editor");
+    if (!g_application_register(G_APPLICATION(application), NULL, &error)) {
+        g_warning("Could not register address book editor: %s", error->message);
+        g_error_free(error);
+    }
     if (g_application_get_is_remote(G_APPLICATION(application))) {
         g_object_unref(application);
         return 1;
@@ -1084,8 +1124,9 @@ main(int argc, char *argv[])
 #endif
 
     bab_init();
-    libbalsa_real_information_func = (LibBalsaInformationFunc)information_real;
-    g_mime_init(GMIME_ENABLE_RFC2047_WORKAROUNDS);
+    balsa_ab_setup_libbalsa_notification((GApplication *) application);
+    g_mime_init();
+    libbalsa_parser_options_init();
 
     /* load address book data */
     libbalsa_conf_push_group("Globals");
@@ -1094,16 +1135,19 @@ main(int argc, char *argv[])
     libbalsa_conf_pop_group();
     libbalsa_conf_foreach_group(ADDRESS_BOOK_SECTION_PREFIX,
                                 bab_config_init, NULL);
-    ab = contacts_app.default_address_book ?
-        contacts_app.default_address_book :
-        contacts_app.address_book_list->data;
+    if (contacts_app.address_book_list == NULL) {
+        /* No address books found--why did we launch the editor? */
+        ab = NULL;
+    } else {
+        ab = contacts_app.default_address_book ?
+            contacts_app.default_address_book :
+            contacts_app.address_book_list->data;
+    }
 
     ab_window = bab_window_new(application);
 
-    g_signal_connect(G_OBJECT(ab_window), "destroy",
+    g_signal_connect(ab_window, "destroy",
                      G_CALLBACK(bab_cleanup), NULL);
-    g_signal_connect(G_OBJECT(ab_window), "delete-event",
-                     G_CALLBACK(bab_delete_ok), NULL);
     bab_set_intial_address_book(ab, ab_window);
 
     /* session management */
