@@ -14,7 +14,7 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -26,11 +26,6 @@
 #include <string.h>
 #include <stdlib.h>
 
-/* for creat(2) */
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-
 #include "filter-funcs.h"
 #include "libbalsa-conf.h"
 #include "misc.h"
@@ -43,22 +38,105 @@
 #  include "macosx-helpers.h"
 #endif
 
+#if HAVE_GCR
+#define GCR_API_SUBJECT_TO_CHANGE
+#include <gcr/gcr.h>
+#else
+#include <gnutls/x509.h>
+#endif
+
 #include <glib/gi18n.h>	/* Must come after balsa-app.h. */
 
 /* Global application structure */
 struct BalsaApplication balsa_app;
 
+#if HAVE_GCR
+
+static gchar *
+ask_password_real(LibBalsaServer *server, const gchar *cert_subject)
+{
+#if defined(HAVE_LIBSECRET)
+    static const gchar *remember_password_message =
+        N_("_Remember password in Secret Service");
+#else
+    static const gchar *remember_password_message =
+        N_("_Remember password");
+#endif                          /* defined(HAVE_LIBSECRET) */
+	GcrPromptDialog *dialog;
+	gchar *prompt;
+	gboolean remember;
+	gchar *passwd;
+        const gchar *user;
+        const gchar *host;
+        const gchar *protocol;
+
+    g_return_val_if_fail(server != NULL, NULL);
+
+    user = libbalsa_server_get_user(server);
+    host = libbalsa_server_get_host(server);
+    protocol = libbalsa_server_get_protocol(server);
+    if (cert_subject != NULL) {
+    	prompt = g_strdup_printf(_("Password to unlock the user certificate\n%s\nfor %s@%s (%s)"),
+    		cert_subject,
+                user, host, protocol);
+    	remember = libbalsa_server_get_remember_cert_passphrase(server);
+    } else {
+    	prompt = g_strdup_printf(_("Password for %s@%s (%s)"),
+                user, host, protocol);
+    	remember = libbalsa_server_get_remember_password(server);
+    }
+	dialog = g_object_new(GCR_TYPE_PROMPT_DIALOG,
+						  "use-header-bar", libbalsa_use_headerbar(),
+						  "title", _("Password needed"),
+						  "description", prompt,
+						  "message", _("Password needed"),
+						  "choice-label", remember_password_message,
+						  "cancel-label", _("_Cancel"),
+						  "continue-label", _("_OK"),
+						  "choice-chosen", remember,
+						  "destroy-with-parent", TRUE,
+						  "transient-for", GTK_WINDOW(balsa_app.main_window),
+						  "modal", TRUE,
+						  NULL);
+	g_free(prompt);
+	passwd = g_strdup(gcr_prompt_password_run(GCR_PROMPT(dialog), NULL, NULL));
+	if (passwd != NULL) {
+		gboolean old_remember;
+
+		old_remember = remember;
+		remember = gcr_prompt_get_choice_chosen(GCR_PROMPT(dialog));
+                if (cert_subject != NULL)
+                    libbalsa_server_set_remember_cert_passphrase(server, remember);
+                else
+                    libbalsa_server_set_remember_password(server, remember);
+		libbalsa_server_set_password(server, passwd, cert_subject != NULL);
+		if (remember || old_remember) {
+			libbalsa_server_config_changed(server);
+		}
+	}
+    gtk_widget_destroy(GTK_WIDGET(dialog));
+    return passwd;
+}
+
+#else
+
 #define HIG_PADDING 12
 
 /* ask_password:
    asks the user for the password to the mailbox on given remote server.
 */
 static gchar *
-ask_password_real(LibBalsaServer * server, LibBalsaMailbox * mbox)
+ask_password_real(LibBalsaServer * server, const gchar *cert_subject)
 {
-    GtkWidget *dialog, *entry, *rememb;
-    GtkWidget *content_area;
-    gchar *prompt, *passwd = NULL;
+	GtkWidget *dialog;
+	GtkWidget *content;
+	GtkWidget *grid;
+	GtkWidget *label;
+    GtkWidget *entry;
+    GtkWidget *rememb_check;
+    gchar *prompt;
+    gchar *passwd;
+	gboolean remember;
 #if defined(HAVE_LIBSECRET)
     static const gchar *remember_password_message =
         N_("_Remember password in Secret Service");
@@ -68,15 +146,14 @@ ask_password_real(LibBalsaServer * server, LibBalsaMailbox * mbox)
 #endif                          /* defined(HAVE_LIBSECRET) */
 
     g_return_val_if_fail(server != NULL, NULL);
-    if (mbox)
-	prompt =
-	    g_strdup_printf(_("Opening remote mailbox %s.\n"
-                              "The _password for %s@%s:"),
-			    mbox->name, server->user, server->host);
-    else
-	prompt =
-	    g_strdup_printf(_("_Password for %s@%s (%s):"), server->user,
-			    server->host, server->protocol);
+    if (cert_subject != NULL) {
+    	prompt = g_strdup_printf(_("Password to unlock the user certificate\n%s\nfor %s@%s (%s)"),
+    		cert_subject, libbalsa_server_get_user(server), libbalsa_server_get_host(server), libbalsa_server_get_protocol(server));
+    	remember = libbalsa_server_get_remember_cert_passphrase(server);
+    } else {
+    	prompt = g_strdup_printf(_("Password for %s@%s (%s)"), libbalsa_server_get_user(server), libbalsa_server_get_host(server), libbalsa_server_get_protocol(server));
+       	remember = libbalsa_server_get_remember_password(server);
+    }
 
     dialog = gtk_dialog_new_with_buttons(_("Password needed"),
                                          GTK_WINDOW(balsa_app.main_window),
@@ -88,43 +165,55 @@ ask_password_real(LibBalsaServer * server, LibBalsaMailbox * mbox)
 #if HAVE_MACOSX_DESKTOP
     libbalsa_macosx_menu_for_parent(dialog, GTK_WINDOW(balsa_app.main_window));
 #endif
-    content_area = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
-    gtk_box_set_spacing(GTK_BOX(content_area), HIG_PADDING);
-    gtk_container_add(GTK_CONTAINER(content_area),
-                      gtk_label_new_with_mnemonic(prompt));
+
+    content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
+    gtk_container_set_border_width(GTK_CONTAINER(content), HIG_PADDING);
+
+    grid = libbalsa_create_grid();
+    gtk_container_add(GTK_CONTAINER(content), grid);
+
+    gtk_grid_attach(GTK_GRID(grid), gtk_label_new(prompt), 0, 0, 2, 1);
     g_free(prompt);
-    gtk_container_add(GTK_CONTAINER(content_area),
-                      entry = gtk_entry_new());
-    gtk_entry_set_width_chars(GTK_ENTRY(entry), 20);
+
+    label = libbalsa_create_grid_label(_("Password:"), grid, 1);
+    entry = libbalsa_create_grid_entry(grid, NULL, NULL, 1, NULL, label);
+    g_object_set(entry, "input-purpose", GTK_INPUT_PURPOSE_PASSWORD, NULL);
     gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
-
-    rememb =  gtk_check_button_new_with_mnemonic(_(remember_password_message));
-    gtk_container_add(GTK_CONTAINER(content_area), rememb);
-    if(server->remember_passwd)
-        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(rememb), TRUE);
-
-    gtk_widget_show_all(content_area);
     gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
-    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
-    gtk_widget_grab_focus (entry);
+    gtk_widget_grab_focus(entry);
 
-    if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
-        unsigned old_rem = server->remember_passwd;
-        passwd = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
-        server->remember_passwd = 
-            !!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(rememb));
-        libbalsa_server_set_password(server, passwd);
-        if( server->remember_passwd || old_rem )
-            libbalsa_server_config_changed(server);
+    rememb_check = libbalsa_create_grid_check(remember_password_message, grid, 2, remember);
+
+    gtk_widget_show_all(grid);
+    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
+
+    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
+		gboolean old_remember;
+
+		old_remember = remember;
+		passwd = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
+		remember = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(rememb_check));
+                if (cert_subject != NULL)
+                    libbalsa_server_set_remember_cert_passphrase(server, remember);
+                else
+                    libbalsa_server_set_remember_password(server, remember);
+		libbalsa_server_set_password(server, passwd, cert_subject != NULL);
+		if (remember || old_remember) {
+			libbalsa_server_config_changed(server);
+		}
+    } else {
+    	passwd = NULL;
     }
     gtk_widget_destroy(dialog);
     return passwd;
 }
 
+#endif
+
 typedef struct {
     GCond cond;
     LibBalsaServer* server;
-    LibBalsaMailbox* mbox;
+    const gchar *cert_subject;
     gchar* res;
     gboolean done;
 } AskPasswdData;
@@ -136,7 +225,7 @@ static gboolean
 ask_passwd_idle(gpointer data)
 {
     AskPasswdData* apd = (AskPasswdData*)data;
-    apd->res = ask_password_real(apd->server, apd->mbox);
+    apd->res = ask_password_real(apd->server, apd->cert_subject);
     apd->done = TRUE;
     g_cond_signal(&apd->cond);
     return FALSE;
@@ -146,7 +235,7 @@ ask_passwd_idle(gpointer data)
    GDK lock must not be held.
 */
 static gchar *
-ask_password_mt(LibBalsaServer * server, LibBalsaMailbox * mbox)
+ask_password_mt(LibBalsaServer * server, const gchar *cert_subject)
 {
     static GMutex ask_passwd_lock;
     AskPasswdData apd;
@@ -154,7 +243,7 @@ ask_password_mt(LibBalsaServer * server, LibBalsaMailbox * mbox)
     g_mutex_lock(&ask_passwd_lock);
     g_cond_init(&apd.cond);
     apd.server = server;
-    apd.mbox   = mbox;
+    apd.cert_subject = cert_subject;
     apd.done   = FALSE;
     g_idle_add(ask_passwd_idle, &apd);
     while (!apd.done) {
@@ -166,84 +255,27 @@ ask_password_mt(LibBalsaServer * server, LibBalsaMailbox * mbox)
     return apd.res;
 }
 
-static gboolean
-set_passwd_from_matching_server(GtkTreeModel *model,
-				GtkTreePath *path,
-				GtkTreeIter *iter,
-				gpointer data)
-{
-    LibBalsaServer *server;
-    LibBalsaServer *master;
-    LibBalsaMailbox *mbox;
-    BalsaMailboxNode *node;
 
-    gtk_tree_model_get(model, iter, 0, &node, -1);
-    g_return_val_if_fail(node != NULL, FALSE);
-    if(node->server) {
-        server = node->server;
-	g_object_unref(node);
-    } else {
-        mbox = node->mailbox;
-	g_object_unref(node);
-        if(!mbox) /* eg. a collection of mboxes */
-            return FALSE;
-        g_return_val_if_fail(LIBBALSA_IS_MAILBOX(mbox), FALSE);
-
-        if (!LIBBALSA_IS_MAILBOX_REMOTE(mbox)) return FALSE;
-        server = LIBBALSA_MAILBOX_REMOTE_SERVER(mbox);
-        g_return_val_if_fail(server != NULL, FALSE);
-    }
-    g_return_val_if_fail(server->host != NULL, FALSE);
-    g_return_val_if_fail(server->user != NULL, FALSE);
-    if (server->passwd == NULL) return FALSE;
-
-    master = (LibBalsaServer *)data;
-    g_return_val_if_fail(LIBBALSA_IS_SERVER(master), FALSE);
-    if (master == server) return FALSE;
-
-    g_return_val_if_fail(server->host != NULL, FALSE);
-    g_return_val_if_fail(server->user != NULL, FALSE);
-
-    if ((strcmp(server->host, master->host) == 0) &&
-	(strcmp(server->user, master->user) == 0)) {
-	g_free(master->passwd);
-	master->passwd = g_strdup(server->passwd);
-	return TRUE;
-    };
-    
-    return FALSE;
-}
 /* ask_password:
    when called from thread, gdk lock must not be held.
+   @param cert_data
 */
 gchar *
-ask_password(LibBalsaServer *server, LibBalsaMailbox *mbox)
+ask_password(LibBalsaServer *server, const gchar *cert_subject, gpointer user_data)
 {
+    G_LOCK_DEFINE_STATIC(ask_password);
     gchar *password;
 
     g_return_val_if_fail(server != NULL, NULL);
-    
-    password = NULL;
-    if (mbox) {
-	gtk_tree_model_foreach(GTK_TREE_MODEL(balsa_app.mblist_tree_store),
-			       (GtkTreeModelForeachFunc)
-			       set_passwd_from_matching_server, server);
 
-	if (server->passwd != NULL) {
-	    password = server->passwd;
-	    server->passwd = NULL;
-	}
+    G_LOCK(ask_password);
+    if (libbalsa_am_i_subthread()) {
+    	password = ask_password_mt(server, cert_subject);
+    } else {
+    	password = ask_password_real(server, cert_subject);
     }
+	G_UNLOCK(ask_password);
 
-    if (!password) {
-        G_LOCK_DEFINE_STATIC(ask_password);
-
-        G_LOCK(ask_password);
-	password = !libbalsa_am_i_subthread() ?
-            ask_password_real(server, mbox) : ask_password_mt(server, mbox);
-        G_UNLOCK(ask_password);
-	return password;
-    }
 	return password;
 }
 
@@ -253,6 +285,10 @@ static gboolean
 send_queued_messages_auto_cb(gpointer data)
 {
 	g_debug("%s: %p", __func__, data);
+
+        if (balsa_app.outbox == NULL)
+            return G_SOURCE_REMOVE;
+
 	libbalsa_process_queue(balsa_app.outbox, balsa_find_sentbox_by_url, balsa_app.smtp_servers, FALSE, NULL);
     return (data == NULL);
 }
@@ -282,7 +318,6 @@ balsa_app_init(void)
     balsa_app.check_mail_auto = TRUE;
     balsa_app.check_mail_timer = 10;
 
-    balsa_app.debug = FALSE;
     balsa_app.previewpane = TRUE;
     balsa_app.pgdownmod = FALSE;
     balsa_app.pgdown_percent = 50;
@@ -290,13 +325,6 @@ balsa_app_init(void)
     /* GUI settings */
     balsa_app.mblist = NULL;
     balsa_app.mblist_width = 100;
-    balsa_app.mw_width = MW_DEFAULT_WIDTH;
-    balsa_app.mw_height = MW_DEFAULT_HEIGHT;
-    balsa_app.mw_maximized = FALSE;
-
-    balsa_app.sw_width = 0;
-    balsa_app.sw_height = 0;
-    balsa_app.sw_maximized = FALSE;
 
     balsa_app.toolbar_wrap_button_text = TRUE;
     balsa_app.send_progress_dialog = TRUE;
@@ -394,11 +422,9 @@ balsa_app_init(void)
     balsa_app.check_imap_inbox = 0;
     balsa_app.imap_scan_depth = 1;
 
-#ifdef HAVE_GPGME
     /* gpgme stuff */
     balsa_app.has_openpgp = FALSE;
     balsa_app.has_smime = FALSE;
-#endif
 
     /* Message filing */
     balsa_app.folder_mru=NULL;
@@ -412,8 +438,7 @@ balsa_app_destroy(void)
 {
     config_save();
 
-    g_list_foreach(balsa_app.address_book_list, (GFunc)g_object_unref, NULL);
-    g_list_free(balsa_app.address_book_list);
+    g_list_free_full(balsa_app.address_book_list, g_object_unref);
     balsa_app.address_book_list = NULL;
 
     /* now free filters */
@@ -422,21 +447,21 @@ balsa_app_destroy(void)
     g_slist_free(balsa_app.filters);
     balsa_app.filters = NULL;
 
-    g_list_foreach(balsa_app.identities, (GFunc)g_object_unref, NULL);
-    g_list_free(balsa_app.identities);
+    g_list_free_full(balsa_app.identities, g_object_unref);
     balsa_app.identities = NULL;
 
+    g_list_free_full(balsa_app.inbox_input, g_object_unref);
+    balsa_app.inbox_input = NULL;
 
-    g_list_foreach(balsa_app.folder_mru, (GFunc)g_free, NULL);
-    g_list_free(balsa_app.folder_mru);
+
+    g_list_free_full(balsa_app.folder_mru, g_free);
     balsa_app.folder_mru = NULL;
 
-    g_list_foreach(balsa_app.fcc_mru, (GFunc)g_free, NULL);
-    g_list_free(balsa_app.fcc_mru);
+    g_list_free_full(balsa_app.fcc_mru, g_free);
     balsa_app.fcc_mru = NULL;
 
 
-    if(balsa_app.debug) g_print("balsa_app: Finished cleaning up.\n");
+    g_debug("balsa_app: Finished cleaning up.");
 }
 
 static gboolean
@@ -444,8 +469,7 @@ check_new_messages_auto_cb(gpointer data)
 {
     check_new_messages_real(balsa_app.main_window, TRUE);
 
-    if (balsa_app.debug)
-        fprintf(stderr, "Auto-checked for new messages…\n");
+    g_debug("Auto-checked for new messages…");
 
     /*  preserver timer */
     return TRUE;
@@ -494,9 +518,7 @@ open_mailbox_by_url(const gchar * url, gboolean hidden)
     LibBalsaMailbox *mailbox;
 
     mailbox = balsa_find_mailbox_by_url(url);
-    if (balsa_app.debug)
-        fprintf(stderr, "balsa_open_mailbox_list: opening %s => %p..\n",
-                url, mailbox);
+    g_debug("balsa_open_mailbox_list: opening %s => %p.", url, mailbox);
     if (mailbox) {
         if (hidden)
             balsa_mblist_open_mailbox_hidden(mailbox);
@@ -544,12 +566,24 @@ balsa_open_mailbox_list(gchar ** urls)
     g_strfreev(urls);
 }
 
+static gint
+position_compare_func(gconstpointer a,
+                      gconstpointer b)
+{
+    const gchar *url_a = *(const gchar **) a;
+    const gchar *url_b = *(const gchar **) b;
+
+    return config_mailbox_get_position(url_a) - config_mailbox_get_position(url_b);
+}
+
 void
 balsa_add_open_mailbox_urls(GPtrArray * url_array)
 {
     libbalsa_conf_foreach_group(VIEW_BY_URL_SECTION_PREFIX,
                                 (LibBalsaConfForeachFunc)
                                 append_url_if_open, url_array);
+
+    g_ptr_array_sort(url_array, position_compare_func);
 }
 
 /* 
@@ -572,7 +606,7 @@ find_mailbox(GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter,
     BalsaMailboxNode *mbnode;
 
     gtk_tree_model_get(model, iter, 0, &mbnode, -1);
-    if (mbnode->mailbox == bf->data) {
+    if (balsa_mailbox_node_get_mailbox(mbnode) == bf->data) {
 	bf->mbnode = mbnode;
 	return TRUE;
     }
@@ -610,8 +644,8 @@ find_path(GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter,
     BalsaMailboxNode *mbnode;
 
     gtk_tree_model_get(model, iter, 0, &mbnode, -1);
-    if (mbnode->server == bf->server &&
-        g_strcmp0(mbnode->dir, bf->data) == 0) {
+    if (balsa_mailbox_node_get_server(mbnode) == bf->server &&
+        g_strcmp0(balsa_mailbox_node_get_dir(mbnode), bf->data) == 0) {
 	bf->mbnode = mbnode;
 	return TRUE;
     }
@@ -642,7 +676,8 @@ find_url(GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter,
     LibBalsaMailbox *mailbox;
 
     gtk_tree_model_get(model, iter, 0, &mbnode, -1);
-    if ((mailbox = mbnode->mailbox) && !strcmp(mailbox->url, bf->data)) {
+    if ((mailbox = balsa_mailbox_node_get_mailbox(mbnode)) != NULL &&
+        strcmp(libbalsa_mailbox_get_url(mailbox), bf->data) == 0) {
         bf->mbnode = mbnode;
         return TRUE;
     }
@@ -691,7 +726,7 @@ balsa_find_mailbox_by_url(const gchar * url)
     LibBalsaMailbox *mailbox = NULL;
 
     if ((mbnode = balsa_find_url(url))) {
-	mailbox = mbnode->mailbox;
+	mailbox = balsa_mailbox_node_get_mailbox(mbnode);
 	g_object_unref(mbnode);
     }
     return mailbox;
@@ -708,16 +743,28 @@ gchar*
 balsa_get_short_mailbox_name(const gchar *url)
 {
     BalsaMailboxNode *mbnode;
+    gchar *short_name = NULL;
 
-    if ((mbnode = balsa_find_url(url)) && mbnode->mailbox) {
-        if (mbnode->server) {
-            return g_strconcat(mbnode->server->host, ":",
-                               mbnode->mailbox->name, NULL);
-        } else {
-            return g_strdup(mbnode->mailbox->name);
+    if ((mbnode = balsa_find_url(url)) != NULL) {
+        LibBalsaMailbox *mailbox = balsa_mailbox_node_get_mailbox(mbnode);
+
+        if (mailbox != NULL) {
+            const gchar *name = libbalsa_mailbox_get_name(mailbox);
+            LibBalsaServer *server = balsa_mailbox_node_get_server(mbnode);
+
+            if (server != NULL) {
+                short_name = g_strconcat(libbalsa_server_get_host(server), ":",
+                                     name, NULL);
+            } else {
+                short_name = g_strdup(name);
+            }
         }
     }
-    return g_strdup(url);
+
+    if (short_name == NULL)
+        short_name = g_strdup(url);
+
+    return short_name;
 }
 
 struct balsa_find_iter_by_data_info {
@@ -736,7 +783,7 @@ balsa_find_iter_by_data_func(GtkTreeModel * model, GtkTreePath * path,
     gtk_tree_model_get(model, iter, 0, &mbnode, -1);
     if(!mbnode)
         return FALSE;
-    if (mbnode == bf->data || mbnode->mailbox == bf->data) {
+    if (mbnode == bf->data || balsa_mailbox_node_get_mailbox(mbnode) == bf->data) {
 	*bf->iter = *iter;
 	bf->found = TRUE;
     }
@@ -783,9 +830,12 @@ ba_remove_children_mailbox_nodes(GtkTreeModel * model, GtkTreeIter * parent,
 	return;
 
     do {
+        LibBalsaMailbox *mailbox;
+
 	gtk_tree_model_get(model, &iter, 0, &mbnode, -1);
-	if (mbnode->parent) {
-	    LibBalsaMailbox *mailbox = mbnode->mailbox;
+	mailbox = balsa_mailbox_node_get_mailbox(mbnode);
+
+	if (balsa_mailbox_node_get_parent(mbnode) != NULL) {
 	    if (mailbox == balsa_app.inbox
 		|| mailbox == balsa_app.outbox
 		|| mailbox == balsa_app.sentbox
@@ -798,12 +848,14 @@ ba_remove_children_mailbox_nodes(GtkTreeModel * model, GtkTreeIter * parent,
 	    valid =
 		gtk_tree_store_remove(balsa_app.mblist_tree_store, &iter);
 	} else {
-	    printf("sparing %s %s\n",
-		   mbnode->mailbox ? "mailbox" : "folder ",
-		   mbnode->mailbox ? mbnode->mailbox->name : mbnode->name);
+	    g_debug("sparing %s %s",
+		   mailbox != NULL ? "mailbox" : "folder ",
+		   mailbox != NULL ?
+                   libbalsa_mailbox_get_name(mailbox) :
+                   balsa_mailbox_node_get_name(mbnode));
 	    valid = gtk_tree_model_iter_next(model, &iter);
 	}
-	g_object_unref(mbnode); 
+	g_object_unref(mbnode);
     } while (valid);
 }
 
@@ -814,10 +866,11 @@ balsa_remove_children_mailbox_nodes(BalsaMailboxNode * mbnode)
     GtkTreeIter parent;
     GtkTreeIter *iter = NULL;
     GSList *specials = NULL, *l;
+    const gchar *name;
 
-    if (balsa_app.debug)
-	printf("Destroying children of %p %s\n",
-	       mbnode, mbnode && mbnode->name ? mbnode->name : "");
+	g_debug("Destroying children of %p %s", mbnode,
+               (mbnode != NULL &&
+                (name = balsa_mailbox_node_get_name(mbnode)) != NULL) ? name : "");
 
     if (mbnode && balsa_find_iter_by_data(&parent, mbnode))
 	iter = &parent;
@@ -839,18 +892,23 @@ BalsaIndex*
 balsa_find_index_by_mailbox(LibBalsaMailbox * mailbox)
 {
     GtkWidget *page;
-    GtkWidget *index;
+    GtkWidget *child;
     guint i;
-    g_return_val_if_fail(balsa_app.notebook, NULL);
+
+    g_return_val_if_fail(LIBBALSA_IS_MAILBOX(mailbox), NULL);
+    g_return_val_if_fail(GTK_IS_NOTEBOOK(balsa_app.notebook), NULL);
 
     for (i = 0;
-	 (page =
-	  gtk_notebook_get_nth_page(GTK_NOTEBOOK(balsa_app.notebook), i));
+	 (page = gtk_notebook_get_nth_page((GtkNotebook *) balsa_app.notebook, i)) != NULL;
 	 i++) {
-        index = gtk_bin_get_child(GTK_BIN(page));
-	if (index && BALSA_INDEX(index)->mailbox_node
-            && BALSA_INDEX(index)->mailbox_node->mailbox == mailbox)
-	    return BALSA_INDEX(index);
+        child = gtk_bin_get_child(GTK_BIN(page));
+	if (child != NULL) {
+            BalsaIndex *bindex = BALSA_INDEX(child);
+            LibBalsaMailbox *this_mailbox = balsa_index_get_mailbox(bindex);
+
+            if (this_mailbox == mailbox)
+                return bindex;
+        }
     }
 
     /* didn't find a matching mailbox */
