@@ -14,7 +14,7 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -41,7 +41,6 @@
    Folders can additionally scan associated directory or IMAP server to
    retrieve their tree of mailboxes.
 */
-static GObjectClass *parent_class = NULL;
 
 typedef struct imap_scan_item_ imap_scan_item;
 struct imap_scan_item_ {
@@ -50,9 +49,6 @@ struct imap_scan_item_ {
     unsigned scanned:1, selectable:1, marked:1;
 };
 
-static void balsa_mailbox_node_class_init(BalsaMailboxNodeClass *
-					     klass);
-static void balsa_mailbox_node_init(BalsaMailboxNode * mn);
 static void balsa_mailbox_node_dispose(GObject * object);
 static void balsa_mailbox_node_finalize(GObject * object);
 
@@ -77,15 +73,13 @@ static BalsaMailboxNode *add_local_folder(BalsaMailboxNode * root,
 					  const char *d_name,
 					  const char *fn);
 
-static void handle_imap_path(const char *fn, char delim, int noselect,
+static void handle_imap_path(const char *fn, int delim, int noselect,
 			     int noscan, int marked, void *data);
 static gint check_imap_path(const char *fn, LibBalsaServer * server,
 			    guint depth);
 static void mark_imap_path(const gchar * fn, gpointer data);
 
 enum {
-    SAVE_CONFIG,
-    LOAD_CONFIG,
     SHOW_PROP_DIALOG,
     APPEND_SUBTREE,
     LAST_SIGNAL
@@ -93,82 +87,47 @@ enum {
 
 static guint balsa_mailbox_node_signals[LAST_SIGNAL];
 
-GType
-balsa_mailbox_node_get_type(void)
-{
-    static GType mailbox_node_type = 0;
+struct _BalsaMailboxNode {
+    GObject object;
+    BalsaMailboxNode *parent; /* NULL for root-level folders & mailboxes */
+    LibBalsaMailbox *mailbox; /* != NULL for leaves only */
+    gchar *name;       /* used for folders, i.e. when mailbox == NULL */
+    time_t last_use;   /* for closing least recently used mailboxes */
+    BalsaMailboxNodeStyle style;
+    /* folder data */
+    gchar* config_prefix;
+    gchar* dir;
+    LibBalsaServer * server; /* Used only by remote; is referenced */
+    int delim; /* IMAP delimiter so that we do not need to check it
+		 * too often. */
 
-    if (!mailbox_node_type) {
-	static const GTypeInfo mailbox_node_info = {
-	    sizeof(BalsaMailboxNodeClass),
-            NULL,               /* base_init */
-            NULL,               /* base_finalize */
-	    (GClassInitFunc)  balsa_mailbox_node_class_init,
-            NULL,               /* class_finalize */
-            NULL,               /* class_data */
-	    sizeof(BalsaMailboxNode),
-            0,                  /* n_preallocs */
-	    (GInstanceInitFunc) balsa_mailbox_node_init
-	};
+    unsigned subscribed:1;     /* Used only by remote */
+    unsigned list_inbox:1;     /* Used only by remote */
+    unsigned scanned:1;        /* IMAP flag */
+};
 
-	mailbox_node_type =
-	    g_type_register_static(G_TYPE_OBJECT, "BalsaMailboxNode",
-                                   &mailbox_node_info, 0);
-    }
-    
-    return mailbox_node_type;
-}
+G_DEFINE_TYPE(BalsaMailboxNode, balsa_mailbox_node, G_TYPE_OBJECT)
 
 static void
 balsa_mailbox_node_class_init(BalsaMailboxNodeClass * klass)
 {
     GObjectClass *object_class;
 
-    parent_class = g_type_class_peek_parent(klass);
-
     object_class = G_OBJECT_CLASS(klass);
 
-    balsa_mailbox_node_signals[SAVE_CONFIG] =
-	g_signal_new("save-config",
-                     G_TYPE_FROM_CLASS(object_class),
-                     G_SIGNAL_RUN_FIRST,
-                     G_STRUCT_OFFSET(BalsaMailboxNodeClass, save_config),
-                     NULL, NULL,
-                     g_cclosure_marshal_VOID__POINTER,
-		     G_TYPE_NONE, 1,
-	             G_TYPE_POINTER);
-    balsa_mailbox_node_signals[LOAD_CONFIG] =
-	g_signal_new("load-config",
-                     G_TYPE_FROM_CLASS(object_class),
-                     G_SIGNAL_RUN_FIRST,
-                     G_STRUCT_OFFSET(BalsaMailboxNodeClass, load_config),
-                     NULL, NULL,
-                     g_cclosure_marshal_VOID__POINTER,
-		     G_TYPE_NONE, 1,
-	             G_TYPE_POINTER);
     balsa_mailbox_node_signals[SHOW_PROP_DIALOG] =
 	g_signal_new("show-prop-dialog",
                      G_TYPE_FROM_CLASS(object_class),
                      G_SIGNAL_RUN_LAST,
-                     G_STRUCT_OFFSET(BalsaMailboxNodeClass,
-                                     show_prop_dialog),
-                     NULL, NULL,
-                     g_cclosure_marshal_VOID__VOID,
+                     0, NULL, NULL, NULL,
 		     G_TYPE_NONE, 0);
     balsa_mailbox_node_signals[APPEND_SUBTREE] =
 	g_signal_new("append-subtree",
                      G_TYPE_FROM_CLASS(object_class),
                      G_SIGNAL_RUN_FIRST,
-                     G_STRUCT_OFFSET(BalsaMailboxNodeClass, append_subtree),
-                     NULL, NULL,
-                     g_cclosure_marshal_VOID__VOID,
+                     0, NULL, NULL, NULL,
 		     G_TYPE_NONE, 0);
 
-    klass->save_config = balsa_mailbox_node_real_save_config;
-    klass->load_config = balsa_mailbox_node_real_load_config;
-    klass->show_prop_dialog = NULL;
-    klass->append_subtree   = NULL;
-
     object_class->dispose = balsa_mailbox_node_dispose;
     object_class->finalize = balsa_mailbox_node_finalize;
 }
@@ -192,17 +151,19 @@ balsa_mailbox_node_dispose(GObject * object)
     BalsaMailboxNode *mn = BALSA_MAILBOX_NODE(object);
     LibBalsaMailbox *mailbox = mn->mailbox;
 
-    if (mailbox) {
-        libbalsa_mailbox_set_open(mailbox,
-                                  libbalsa_mailbox_is_open(mailbox));
-        config_save_mailbox_view(mailbox->url, mailbox->view);
-	if (balsa_app.main_window)
-	    balsa_window_close_mbnode(balsa_app.main_window, mn);
+    if (mailbox != NULL) {
+        if (!LIBBALSA_IS_MAILBOX_POP3(mailbox)) {
+            libbalsa_mailbox_set_open(mailbox, libbalsa_mailbox_is_open(mailbox));
+            config_save_mailbox_view(libbalsa_mailbox_get_url(mailbox),
+                                     libbalsa_mailbox_get_view(mailbox));
+            if (balsa_app.main_window != NULL)
+                balsa_window_close_mbnode(balsa_app.main_window, mn);
+        }
 	g_object_unref(mailbox);
 	mn->mailbox = NULL;
     }
 
-    G_OBJECT_CLASS(parent_class)->dispose(object);
+    G_OBJECT_CLASS(balsa_mailbox_node_parent_class)->dispose(object);
 }
 
 static void
@@ -225,7 +186,7 @@ balsa_mailbox_node_finalize(GObject * object)
 	mn->server = NULL;
     }
 
-    G_OBJECT_CLASS(parent_class)->finalize(G_OBJECT(object));
+    G_OBJECT_CLASS(balsa_mailbox_node_parent_class)->finalize(object);
 }
 
 static void
@@ -236,14 +197,14 @@ balsa_mailboxes_append_children(GtkTreeModel *model, GtkTreeIter *parent,
 
     if (!gtk_tree_model_iter_children(model, &child, parent))
 	return;
+
     do {
         BalsaMailboxNode *mbnode;
         gchar *str;
 
         gtk_tree_model_get(model, &child, 0, &mbnode, -1);
-        g_return_if_fail(mbnode->server);
-        if(mbnode->mailbox) {
-            const char *path = 
+        if (mbnode->mailbox != NULL) {
+            const char *path =
                 libbalsa_mailbox_imap_get_path
                 (LIBBALSA_MAILBOX_IMAP(mbnode->mailbox));
 
@@ -257,35 +218,32 @@ balsa_mailboxes_append_children(GtkTreeModel *model, GtkTreeIter *parent,
         g_object_unref(mbnode);
 
         balsa_mailboxes_append_children(model, &child, children_names);
-    } while(gtk_tree_model_iter_next(model, &child));
+    } while (gtk_tree_model_iter_next(model, &child));
 }
 
 static void
 balsa_mailbox_node_real_save_config(BalsaMailboxNode* mn, const gchar * group)
 {
-    GPtrArray *children_names;
     GtkTreeIter iter;
-    
-    g_return_if_fail(!mn->parent);
 
-    if(mn->name)
-	printf("Saving mailbox node %s with group %s\n", mn->name, group);
+    if(mn->name != NULL)
+	g_debug("Saving mailbox node %s with group %s", mn->name, group);
     libbalsa_imap_server_save_config(LIBBALSA_IMAP_SERVER(mn->server));
     libbalsa_conf_set_string("Name",      mn->name);
     libbalsa_conf_set_string("Directory", mn->dir);
     libbalsa_conf_set_bool("Subscribed",  mn->subscribed);
     libbalsa_conf_set_bool("ListInbox",   mn->list_inbox);
-    
+
     if (balsa_find_iter_by_data(&iter, mn)) {
+        GPtrArray *children_names;
         GtkTreeModel *model = GTK_TREE_MODEL(balsa_app.mblist_tree_store);
 
-        children_names = g_ptr_array_new();
+        children_names = g_ptr_array_new_with_free_func(g_free);
         balsa_mailboxes_append_children(model, &iter,
                                         children_names);
-        printf("Saving %d children\n", children_names->len);
+        g_debug("Saving %d children", children_names->len);
         libbalsa_conf_set_vector("Children", children_names->len,
                                  (const char*const*)(children_names->pdata));
-        g_ptr_array_foreach(children_names, (GFunc)g_free, NULL);
         g_ptr_array_free(children_names, TRUE);
     }
 
@@ -366,9 +324,7 @@ check_local_path(const gchar * path, guint depth)
     libbalsa_conf_foreach_group(VIEW_BY_URL_SECTION_PREFIX,
                                 (LibBalsaConfForeachFunc) check_url_func,
                                 &cpi);
-    if(balsa_app.debug) 
-	printf("check_local_path: path “%s” must_scan %d.\n",
-               cpi.url, cpi.must_scan);
+    g_debug("check_local_path: path “%s” must_scan %d.", cpi.url, cpi.must_scan);
     g_free(cpi.url);
 
     return cpi.must_scan;
@@ -403,8 +359,10 @@ load_mailbox_view(BalsaMailboxNode * mbnode)
 {
     LibBalsaMailbox *mailbox = mbnode->mailbox;
 
-    if (!mailbox->view)
-        mailbox->view = config_load_mailbox_view(mailbox->url);
+    if (!libbalsa_mailbox_get_view(mailbox))
+        libbalsa_mailbox_set_view(mailbox,
+                                  config_load_mailbox_view
+                                  (libbalsa_mailbox_get_url(mailbox)));
 }
 
 static gboolean
@@ -415,10 +373,10 @@ imap_scan_attach_mailbox(BalsaMailboxNode * mbnode, imap_scan_item * isi)
 
     /* If the mailbox was added from the config file, it is already
      * connected to "show-prop-dialog". */
-    if (!g_signal_has_handler_pending(G_OBJECT(mbnode),
+    if (!g_signal_has_handler_pending(mbnode,
                                       balsa_mailbox_node_signals
                                       [SHOW_PROP_DIALOG], 0, FALSE))
-	g_signal_connect(G_OBJECT(mbnode), "show-prop-dialog",
+	g_signal_connect(mbnode, "show-prop-dialog",
                          G_CALLBACK(folder_conf_imap_sub_node), NULL);
     if (LIBBALSA_IS_MAILBOX_IMAP(mbnode->mailbox))
         /* it already has a mailbox */
@@ -429,22 +387,21 @@ imap_scan_attach_mailbox(BalsaMailboxNode * mbnode, imap_scan_item * isi)
     libbalsa_mailbox_remote_set_server(LIBBALSA_MAILBOX_REMOTE(m),
 				       mbnode->server);
     libbalsa_mailbox_imap_set_path(m, isi->fn);
-    if(balsa_app.debug)
-        printf("imap_scan_attach_mailbox: add mbox of name %s "
-	       "(full path %s)\n", isi->fn, mailbox->url);
-    /* avoid allocating the name again: */
-    mailbox->name = mbnode->name;
+    g_debug("imap_scan_attach_mailbox: add mbox of name %s (full path %s)", isi->fn,
+            libbalsa_mailbox_get_url(mailbox));
+    libbalsa_mailbox_set_name(mailbox, mbnode->name);
+    g_free(mbnode->name);
     mbnode->name = NULL;
     mbnode->mailbox = mailbox;
     load_mailbox_view(mbnode);
     if (isi->special) {
 	if (*isi->special)
 	    g_object_remove_weak_pointer(G_OBJECT(*isi->special),
-					 (gpointer) isi->special);
+					 (gpointer *) isi->special);
         *isi->special = mailbox;
-	g_object_add_weak_pointer(G_OBJECT(m), (gpointer) isi->special);
+	g_object_add_weak_pointer(G_OBJECT(m), (gpointer *) isi->special);
         if (isi->special == &balsa_app.outbox)
-            mailbox->no_reassemble = TRUE;
+            libbalsa_mailbox_set_no_reassemble(mailbox, TRUE);
     }
 
     return TRUE;
@@ -466,7 +423,7 @@ imap_scan_attach_mailbox(BalsaMailboxNode * mbnode, imap_scan_item * isi)
 typedef struct imap_scan_tree_ imap_scan_tree;
 struct imap_scan_tree_ {
     GSList *list;               /* a list of imap_scan_items */
-    char delim;
+    int delim;
 };
 
 static void imap_scan_destroy_tree(imap_scan_tree * tree);
@@ -486,7 +443,7 @@ imap_dir_cb(BalsaMailboxNode* mb)
     if (!balsa_app.main_window)
         return;
 
-    statusbar = GTK_STATUSBAR(balsa_app.main_window->statusbar);
+    statusbar = balsa_window_get_statusbar(balsa_app.main_window);
     context_id = gtk_statusbar_get_context_id(statusbar, "MailboxNode imap");
 
     if(restore_children_from_cache(mb))
@@ -513,7 +470,7 @@ imap_dir_cb(BalsaMailboxNode* mb)
                              ? _("Scanning of %s failed: %s\n"
                                  "Check network connectivity.")
                              : _("Scanning of %s failed: %s"),
-                             mb->server->host,
+                             libbalsa_server_get_host(mb->server),
                              error->message);
         g_error_free(error);
         imap_scan_destroy_tree(&imap_tree);
@@ -541,10 +498,10 @@ imap_dir_cb(BalsaMailboxNode* mb)
     }
     imap_scan_destroy_tree(&imap_tree);
 
-    if(balsa_app.debug && mb->name)
-        printf("imap_dir_cb:  main mailbox node %s mailbox is %p\n", 
-               mb->name, mb->mailbox);
-    if(balsa_app.debug) printf("%d: Scanning done.\n", (int)time(NULL));
+    if (mb->name) {
+    	g_debug("imap_dir_cb: main mailbox node %s mailbox is %p", mb->name, mb->mailbox);
+    }
+    g_debug("%d: Scanning done.", (int)time(NULL));
     gtk_statusbar_pop(statusbar, context_id);
 
     /* We can save the cache now... */
@@ -558,14 +515,15 @@ balsa_mailbox_node_new_from_mailbox(LibBalsaMailbox * mb)
     mbn = BALSA_MAILBOX_NODE(balsa_mailbox_node_new());
     mbn->mailbox = mb;
     load_mailbox_view(mbn);
-    g_signal_connect(G_OBJECT(mbn), "show-prop-dialog", 
+    g_signal_connect(mbn, "show-prop-dialog", 
 		     G_CALLBACK(mailbox_conf_edit), NULL);
     if (LIBBALSA_IS_MAILBOX_MH(mb) || LIBBALSA_IS_MAILBOX_MAILDIR(mb)) {
 	/* Mh and Maildir mailboxes are directories, and may be nested,
 	 * so we need to be able to append a subtree. */
-	mbn->name = g_strdup(libbalsa_mailbox_local_get_path(mb));
+	mbn->name =
+            g_strdup(libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mb));
 	mbn->dir = g_strdup(mbn->name);
-	g_signal_connect(G_OBJECT(mbn), "append-subtree", 
+	g_signal_connect(mbn, "append-subtree", 
                          G_CALLBACK(read_dir_cb), NULL);
     }
     return mbn;
@@ -580,9 +538,9 @@ balsa_mailbox_node_new_from_dir(const gchar* dir)
 
     mbn->name = g_strdup(dir);
     mbn->dir  = g_strdup(dir);
-    g_signal_connect(G_OBJECT(mbn), "show-prop-dialog", 
+    g_signal_connect(mbn, "show-prop-dialog", 
 		     G_CALLBACK(dir_conf_edit), NULL);
-    g_signal_connect(G_OBJECT(mbn), "append-subtree", 
+    g_signal_connect(mbn, "append-subtree", 
 		     G_CALLBACK(read_dir_cb), NULL);
     return mbn;
 }
@@ -605,20 +563,22 @@ balsa_mailbox_node_new_from_config(const gchar* group)
 
     folder->server = LIBBALSA_SERVER(libbalsa_imap_server_new_from_config());
 
-    if(balsa_app.debug)
-	printf("Server loaded, host: %s, %s.\n", folder->server->host,
-	       folder->server->use_ssl ? "SSL" : "no SSL");
-    g_signal_connect_swapped(G_OBJECT(folder->server), "config-changed", 
+	g_debug("Server loaded, host: %s, security %d.", libbalsa_server_get_host(folder->server),
+		libbalsa_server_get_security(folder->server));
+    g_signal_connect_swapped(folder->server, "config-changed", 
                              G_CALLBACK(config_folder_update), folder);
-    g_signal_connect(G_OBJECT(folder), "show-prop-dialog", 
+    g_signal_connect(folder, "show-prop-dialog", 
 		     G_CALLBACK(folder_conf_imap_node), NULL);
-    g_signal_connect(G_OBJECT(folder), "append-subtree", 
+    g_signal_connect(folder, "append-subtree", 
 		     G_CALLBACK(imap_dir_cb), NULL);
     libbalsa_server_connect_signals(folder->server,
                                     G_CALLBACK(ask_password), NULL);
     balsa_mailbox_node_load_config(folder, group);
 
     folder->dir = libbalsa_conf_get_string("Directory");
+    if (folder->dir == NULL) {
+    	folder->dir = g_strdup("");
+    }
     folder->subscribed =
 	libbalsa_conf_get_bool("Subscribed"); 
     folder->list_inbox =
@@ -643,35 +603,19 @@ balsa_mailbox_node_new_imap_node(LibBalsaServer* s, const char*p)
 
     folder->server = s;
     folder->dir = g_strdup(p);
-    g_signal_connect(G_OBJECT(folder), "append-subtree", 
+    g_signal_connect(folder, "append-subtree", 
 		     G_CALLBACK(imap_dir_cb), NULL);
 
     return folder;
 }
 
-BalsaMailboxNode*
-balsa_mailbox_node_new_imap(LibBalsaServer* s, const char*p)
-{
-    BalsaMailboxNode * folder = balsa_mailbox_node_new_imap_node(s, p);
-    g_assert(s);
-
-    folder->mailbox = libbalsa_mailbox_imap_new();
-    g_object_ref(G_OBJECT(folder->mailbox));
-    libbalsa_mailbox_remote_set_server(
-	LIBBALSA_MAILBOX_REMOTE(folder->mailbox), s);
-    libbalsa_mailbox_imap_set_path(LIBBALSA_MAILBOX_IMAP(folder->mailbox), p);
-    load_mailbox_view(folder);
-
-    return folder;
-}
-
 BalsaMailboxNode*
 balsa_mailbox_node_new_imap_folder(LibBalsaServer* s, const char*p)
 {
     BalsaMailboxNode * folder = balsa_mailbox_node_new_imap_node(s, p);
     g_assert(s);
 
-    g_signal_connect(G_OBJECT(folder), "show-prop-dialog", 
+    g_signal_connect(folder, "show-prop-dialog", 
 		     G_CALLBACK(folder_conf_imap_node), NULL);
     return folder;
 }
@@ -679,19 +623,20 @@ balsa_mailbox_node_new_imap_folder(LibBalsaServer* s, const char*p)
 void
 balsa_mailbox_node_show_prop_dialog(BalsaMailboxNode* mn)
 {
-    if (mn)
-        g_signal_emit(G_OBJECT(mn),
-                      balsa_mailbox_node_signals[SHOW_PROP_DIALOG], 0);
+    if (mn != NULL)
+        g_signal_emit(mn, balsa_mailbox_node_signals[SHOW_PROP_DIALOG], 0);
 }
 
 void
 balsa_mailbox_node_append_subtree(BalsaMailboxNode * mn)
 {
-    g_signal_emit(G_OBJECT(mn),
-		  balsa_mailbox_node_signals[APPEND_SUBTREE], 0);
+    if (!mn->scanned) {
+        g_signal_emit(mn, balsa_mailbox_node_signals[APPEND_SUBTREE], 0);
+        mn->scanned = TRUE;
+    }
 }
 
-void 
+void
 balsa_mailbox_node_show_prop_dialog_cb(GtkWidget * widget, gpointer data)
 {
     balsa_mailbox_node_show_prop_dialog((BalsaMailboxNode*)data);
@@ -704,15 +649,13 @@ balsa_mailbox_node_show_prop_dialog_cb(GtkWidget * widget, gpointer data)
 void
 balsa_mailbox_node_load_config(BalsaMailboxNode* mn, const gchar* group)
 {
-    g_signal_emit(G_OBJECT(mn),
-		  balsa_mailbox_node_signals[LOAD_CONFIG], 0, group);
+    balsa_mailbox_node_real_load_config(mn, group);
 }
 
 void
 balsa_mailbox_node_save_config(BalsaMailboxNode* mn, const gchar* group)
 {
-    g_signal_emit(G_OBJECT(mn),
-		  balsa_mailbox_node_signals[SAVE_CONFIG], 0, group);
+    balsa_mailbox_node_real_save_config(mn, group);
 }
 
 /* ---------------------------------------------------------------------
@@ -735,7 +678,7 @@ balsa_mailbox_local_append(LibBalsaMailbox* mbx)
 
     g_return_if_fail(LIBBALSA_IS_MAILBOX_LOCAL(mbx));
 
-    dir = g_strdup(libbalsa_mailbox_local_get_path(mbx));
+    dir = g_strdup(libbalsa_mailbox_local_get_path((LibBalsaMailboxLocal *) mbx));
     while (dir[1] /* i.e. dir != "/" */
            && !(parent = find_dir(dir))) {
         gchar* tmp =  g_path_get_dirname(dir);
@@ -795,8 +738,9 @@ restore_children_from_cache(BalsaMailboxNode *mn)
     imap_scan_item isi;
 
     if(!children) {
-        if(mn->name)
-            printf("No cache for %s - quitting.\n", mn->name);
+        if (mn->name) {
+        	g_debug("No cache for %s - quitting.", mn->name);
+        }
         return FALSE;
     }
 
@@ -847,13 +791,13 @@ balsa_mailbox_node_scan_children(BalsaMailboxNode * mbnode)
 	    g_object_unref(mn);
         }
     } else
-        g_print("balsa_mailbox_node_scan_children: didn't find mbnode.\n");
+        g_warning("balsa_mailbox_node_scan_children: didn't find mbnode.");
 
     if (list && !g_object_get_data(G_OBJECT(mbnode),
                                    BALSA_MAILBOX_NODE_LIST_KEY)) {
         BalsaMailboxNode **mn = g_new(BalsaMailboxNode *, 1);
         *mn = mbnode;
-        g_object_add_weak_pointer(G_OBJECT(mbnode), (gpointer) mn);
+        g_object_add_weak_pointer(G_OBJECT(mbnode), (gpointer *) mn);
         g_object_set_data(G_OBJECT(mbnode), BALSA_MAILBOX_NODE_LIST_KEY,
                           g_slist_reverse(list));
         g_idle_add((GSourceFunc) bmbn_scan_children_idle, mn);
@@ -882,17 +826,20 @@ bmbn_scan_children_idle(BalsaMailboxNode ** mbnode)
             gboolean has_unread_messages = FALSE;
 
             if (mn->mailbox)
-                has_unread_messages = mn->mailbox->has_unread_messages;
+                has_unread_messages =
+                    libbalsa_mailbox_get_has_unread_messages(mn->mailbox);
             balsa_mailbox_node_rescan(mn);
             if (!l->data)
                 continue;
             if (mn->mailbox)
-                mn->mailbox->has_unread_messages = has_unread_messages;
+                libbalsa_mailbox_set_has_unread_messages
+                    (mn->mailbox, has_unread_messages);
             mn->scanned = TRUE;
-        } else if (balsa_app.debug)
-            g_print("%s: %s “%s” was already scanned\n", __func__,
+        } else {
+            g_debug("%s: %s “%s” was already scanned", __func__,
                     mn->mailbox ? "mailbox" : "folder",
-                    mn->mailbox ? mn->mailbox->name : mn->name);
+                    mn->mailbox ? libbalsa_mailbox_get_name(mn->mailbox) : mn->name);
+        }
         g_object_remove_weak_pointer(G_OBJECT(mn), & l->data);
     }
     g_slist_free(list);
@@ -900,7 +847,7 @@ bmbn_scan_children_idle(BalsaMailboxNode ** mbnode)
     if (*mbnode) {
         g_object_set_data(G_OBJECT(*mbnode), BALSA_MAILBOX_NODE_LIST_KEY,
                           NULL);
-        g_object_remove_weak_pointer(G_OBJECT(*mbnode), (gpointer) mbnode);
+        g_object_remove_weak_pointer(G_OBJECT(*mbnode), (gpointer *) mbnode);
     }
     g_free(mbnode);
 
@@ -917,10 +864,10 @@ add_menu_entry(GtkWidget * menu, const gchar * label, GCallback cb,
     GtkWidget *menuitem;
 
     menuitem = label ? gtk_menu_item_new_with_mnemonic(label)
-	: gtk_menu_item_new();
+	: gtk_separator_menu_item_new();
 
     if (cb)
-	g_signal_connect(G_OBJECT(menuitem), "activate",
+	g_signal_connect(menuitem, "activate",
 			 G_CALLBACK(cb), mbnode);
 
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
@@ -983,22 +930,6 @@ mb_draftbox_cb(GtkWidget * widget, BalsaMailboxNode * mbnode)
     config_mailbox_set_as_special(mbnode->mailbox, SPECIAL_DRAFT);
 }
 
-static void
-mb_subscribe_cb(GtkWidget * widget, BalsaMailboxNode * mbnode)
-{
-    g_return_if_fail(LIBBALSA_IS_MAILBOX_IMAP(mbnode->mailbox));
-    libbalsa_mailbox_imap_subscribe(LIBBALSA_MAILBOX_IMAP(mbnode->mailbox), 
-				    TRUE);
-}
-
-static void
-mb_unsubscribe_cb(GtkWidget * widget, BalsaMailboxNode * mbnode)
-{
-    g_return_if_fail(LIBBALSA_IS_MAILBOX_IMAP(mbnode->mailbox));
-    libbalsa_mailbox_imap_subscribe(LIBBALSA_MAILBOX_IMAP(mbnode->mailbox),
-				     FALSE);
-}
-
 static void
 mb_rescan_cb(GtkWidget * widget, BalsaMailboxNode * mbnode)
 {
@@ -1029,7 +960,8 @@ mb_filter_cb(GtkWidget * widget, BalsaMailboxNode * mbnode)
 	   problems of infinite recursion (when one mailbox being
 	   filtered is also the destination of the filter action (eg a
 	   copy)). So let's see that later :) */
-	g_print("You can apply filters only on mailbox\n");
+    	libbalsa_information_parented(GTK_WINDOW(balsa_app.main_window),
+    		LIBBALSA_INFORMATION_MESSAGE, _("You can apply filters only on mailbox"));
 }
 
 static void
@@ -1038,6 +970,14 @@ mb_empty_trash_cb(GtkWidget * widget, BalsaMailboxNode * mbnode)
     empty_trash(balsa_app.main_window);
 }
 
+gboolean
+balsa_mailbox_node_is_imap(const BalsaMailboxNode *mbnode)
+{
+	return (mbnode != NULL) &&
+		(mbnode->server != NULL) &&
+		(strcmp(libbalsa_server_get_protocol(mbnode->server), "imap") == 0);
+}
+
 GtkWidget *
 balsa_mailbox_node_get_context_menu(BalsaMailboxNode * mbnode)
 {
@@ -1051,7 +991,7 @@ balsa_mailbox_node_get_context_menu(BalsaMailboxNode * mbnode)
 
     menu = gtk_menu_new();
     /* it's a single-use menu, so we must destroy it when we're done */
-    g_signal_connect(G_OBJECT(menu), "selection-done",
+    g_signal_connect(menu, "selection-done",
                      G_CALLBACK(gtk_widget_destroy), NULL);
 
     submenu = gtk_menu_new();
@@ -1061,13 +1001,13 @@ balsa_mailbox_node_get_context_menu(BalsaMailboxNode * mbnode)
 		   G_CALLBACK(mailbox_conf_add_maildir_cb), NULL);
     add_menu_entry(submenu, _("Local M_H mailbox…"),
 		   G_CALLBACK(mailbox_conf_add_mh_cb), NULL);
-    add_menu_entry(submenu, _("Remote _IMAP mailbox…"), 
-		   G_CALLBACK(mailbox_conf_add_imap_cb), NULL);
     add_menu_entry(submenu, NULL, NULL, mbnode);
     add_menu_entry(submenu, _("Remote IMAP _folder…"), 
 		   G_CALLBACK(folder_conf_add_imap_cb), NULL);
-    add_menu_entry(submenu, _("Remote IMAP _subfolder…"), 
-		   G_CALLBACK(folder_conf_add_imap_sub_cb), NULL);
+    if (balsa_mailbox_node_is_imap(mbnode)) {
+    	add_menu_entry(submenu, _("Remote IMAP _subfolder…"),
+		   G_CALLBACK(folder_conf_add_imap_sub_cb), mbnode);
+    }
     gtk_widget_show(submenu);
     
     /* Translators: popup menu item "New" mailbox or folder */
@@ -1085,13 +1025,13 @@ balsa_mailbox_node_get_context_menu(BalsaMailboxNode * mbnode)
     /* If we didn't click on a mailbox node then there is only one option. */
     add_menu_entry(menu, NULL, NULL, NULL);
 
-    if (g_signal_has_handler_pending(G_OBJECT(mbnode),
+    if (g_signal_has_handler_pending(mbnode,
                                      balsa_mailbox_node_signals
                                      [SHOW_PROP_DIALOG], 0, FALSE))
         add_menu_entry(menu, _("_Properties…"),
                        G_CALLBACK(mb_conf_cb), mbnode);
 
-    if (g_signal_has_handler_pending(G_OBJECT(mbnode),
+    if (g_signal_has_handler_pending(mbnode,
 		                     balsa_mailbox_node_signals
 				     [APPEND_SUBTREE], 0, FALSE))
 	add_menu_entry(menu, _("_Rescan"),
@@ -1116,14 +1056,6 @@ balsa_mailbox_node_get_context_menu(BalsaMailboxNode * mbnode)
     if (!special && !mbnode->config_prefix)
 	add_menu_entry(menu, _("_Delete"), G_CALLBACK(mb_del_cb),  mbnode);
 
-    if(LIBBALSA_IS_MAILBOX_IMAP(mailbox)) {
-        add_menu_entry(menu, NULL, NULL, NULL);
-        add_menu_entry(menu, _("_Subscribe"),   
-                       G_CALLBACK(mb_subscribe_cb),   mbnode);
-        add_menu_entry(menu, _("_Unsubscribe"), 
-                       G_CALLBACK(mb_unsubscribe_cb), mbnode);
-    }
-
     if (!special) {
         add_menu_entry(menu, NULL, NULL, NULL);
         add_menu_entry(menu, _("Mark as _Inbox"),    
@@ -1172,28 +1104,28 @@ remove_mailbox_from_nodes(LibBalsaMailbox* mailbox)
 	mbnode = balsa_mailbox_node_new_from_mailbox(mailbox);
     return mbnode;
 }
-    
-static BalsaMailboxNode * 
+
+static BalsaMailboxNode *
 remove_special_mailbox_by_url(const gchar* url, LibBalsaMailbox *** special)
 {
     LibBalsaMailbox **mailbox;
 
-    if (g_strcmp0(url, balsa_app.trash->url) == 0)
+    if (g_strcmp0(url, libbalsa_mailbox_get_url(balsa_app.trash)) == 0)
 	mailbox = &balsa_app.trash;
-    else if (g_strcmp0(url, balsa_app.inbox->url) == 0)
+    else if (g_strcmp0(url, libbalsa_mailbox_get_url(balsa_app.inbox)) == 0)
 	mailbox = &balsa_app.inbox;
-    else if (g_strcmp0(url, balsa_app.outbox->url) == 0)
+    else if (g_strcmp0(url, libbalsa_mailbox_get_url(balsa_app.outbox)) == 0)
 	mailbox = &balsa_app.outbox;
-    else if (g_strcmp0(url, balsa_app.sentbox->url) == 0)
+    else if (g_strcmp0(url, libbalsa_mailbox_get_url(balsa_app.sentbox)) == 0)
 	mailbox = &balsa_app.sentbox;
-    else if (g_strcmp0(url, balsa_app.draftbox->url) == 0)
+    else if (g_strcmp0(url, libbalsa_mailbox_get_url(balsa_app.draftbox)) == 0)
 	mailbox = &balsa_app.draftbox;
     else
         mailbox = NULL;
 
     if (special)
         *special = mailbox;
-	
+
     return mailbox ? remove_mailbox_from_nodes(*mailbox) : NULL;
 }
 
@@ -1232,16 +1164,15 @@ add_local_mailbox(BalsaMailboxNode *root, const gchar * name,
 	    mailbox = NULL;
 	}
 	if(!mailbox) {/* local mailbox could not be created; privileges? */
-	    printf("Not accessible mailbox %s\n", path);
+		g_warning("Not accessible mailbox %s", path);
 	    return NULL;
 	}
-	mailbox->name = g_strdup(name);
+	libbalsa_mailbox_set_name(mailbox, name);
 	
 	mbnode = balsa_mailbox_node_new_from_mailbox(mailbox);
 	
-	if (balsa_app.debug)
-	    g_print(_("Local mailbox %s loaded as: %s\n"),
-		    mailbox->name,
+	    g_debug("Local mailbox %s loaded as: %s",
+		    libbalsa_mailbox_get_name(mailbox),
 		    g_type_name(G_OBJECT_TYPE(mailbox)));
         if (balsa_app.check_mail_upon_startup
             && libbalsa_mailbox_get_subscribe(mailbox) !=
@@ -1270,8 +1201,7 @@ add_local_folder(BalsaMailboxNode * root, const char *d_name,
     mbnode = balsa_mailbox_node_new_from_dir(path);
     mbnode->parent = root;
     balsa_mblist_mailbox_node_append(root, mbnode);
-    if (balsa_app.debug)
-	g_print(_("Local folder %s\n"), path);
+	g_debug("Local folder %s", path);
 
     return mbnode;
 }
@@ -1281,7 +1211,7 @@ add_local_folder(BalsaMailboxNode * root, const char *d_name,
  * IMAP folder scanner functions 
  * --------------------------------------------------------------------- */
 static gchar*
-get_parent_folder_name(const gchar* path, char delim)
+get_parent_folder_name(const gchar* path, gint delim)
 {
     const gchar* last_delim = strrchr(path, delim);
     return last_delim ? g_strndup(path, last_delim-path)
@@ -1361,7 +1291,7 @@ imap_scan_create_mbnode(BalsaMailboxNode * root, imap_scan_item * isi,
    add given mailbox unless its base name begins on dot.
 */
 static void
-add_imap_entry(const char *fn, char delim, gboolean noscan,
+add_imap_entry(const char *fn, int delim, gboolean noscan,
 	       gboolean selectable, gboolean marked, void *data)
 {
     imap_scan_tree *tree = (imap_scan_tree *) data;
@@ -1392,7 +1322,7 @@ imap_scan_destroy_tree(imap_scan_tree * tree)
 }
 
 static void
-handle_imap_path(const char *fn, char delim, int noselect, int noscan,
+handle_imap_path(const char *fn, int delim, int noselect, int noscan,
 		 int marked, void *data)
 {
     if (!noselect) {
@@ -1401,10 +1331,8 @@ handle_imap_path(const char *fn, char delim, int noselect, int noscan,
 	    /* ignore mailboxes that begin with a dot */
 	    return;
     }
-    if (balsa_app.debug)
-	printf("handle_imap_path: Adding mailbox of path “%s” "
-	       "delim `%c' noselect %d noscan %d\n",
-	       fn, delim, noselect, noscan);
+    g_debug("handle_imap_path: Adding mailbox of path “%s” delim `%c' noselect %d noscan %d",
+    	fn, delim, noselect, noscan);
     add_imap_entry(fn, delim, noscan, !noselect, marked, data);
 }
 
@@ -1431,9 +1359,7 @@ check_imap_path(const gchar *fn, LibBalsaServer * server, guint depth)
     libbalsa_conf_foreach_group(VIEW_BY_URL_SECTION_PREFIX,
                                 (LibBalsaConfForeachFunc) check_url_func,
                                 &cpi);
-    if(balsa_app.debug) 
-	printf("check_imap_path: path “%s” must_scan %d.\n",
-               cpi.url, cpi.must_scan);
+    g_debug("check_imap_path: path “%s” must_scan %d.", cpi.url, cpi.must_scan);
     g_free(cpi.url);
 
     return cpi.must_scan;
@@ -1449,8 +1375,7 @@ mark_imap_path(const gchar * fn, gpointer data)
     imap_scan_tree *tree = data;
     GSList *list;
 
-    if(balsa_app.debug) 
-	printf("mark_imap_path: find path “%s”.\n", fn);
+    g_debug("mark_imap_path: find path “%s”.", fn);
     for (list = tree->list; list; list = list->next) {
         imap_scan_item *item = list->data;
         if (!strcmp(item->fn, fn)) {
@@ -1458,6 +1383,165 @@ mark_imap_path(const gchar * fn, gpointer data)
             break;
         }
     }
-    if (!list && balsa_app.debug)
-	printf(" not found.\n");
+    if (!list) {
+    	g_debug("… not found.");
+    }
+}
+
+/*
+ * Setters
+ */
+
+void
+balsa_mailbox_node_set_dir(BalsaMailboxNode * mbnode, const gchar * dir)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    g_free(mbnode->dir);
+    mbnode->dir = g_strdup(dir);
+}
+
+void
+balsa_mailbox_node_set_name(BalsaMailboxNode * mbnode, const gchar * name)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    g_free(mbnode->name);
+    mbnode->name = g_strdup(name);
+}
+
+void
+balsa_mailbox_node_set_config_prefix(BalsaMailboxNode * mbnode, const gchar * config_prefix)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    g_free(mbnode->config_prefix);
+    mbnode->config_prefix = g_strdup(config_prefix);
+}
+
+void
+balsa_mailbox_node_set_last_use_time(BalsaMailboxNode * mbnode)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    time(&mbnode->last_use);
+}
+
+void
+balsa_mailbox_node_change_style(BalsaMailboxNode * mbnode,
+                                BalsaMailboxNodeStyle set,
+                                BalsaMailboxNodeStyle clear)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    mbnode->style |= set;
+    mbnode->style &= ~clear;
+}
+
+void
+balsa_mailbox_node_set_list_inbox(BalsaMailboxNode * mbnode, guint list_inbox)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    mbnode->list_inbox = !!list_inbox;
+}
+
+void
+balsa_mailbox_node_set_subscribed(BalsaMailboxNode * mbnode, guint subscribed)
+{
+    g_return_if_fail(BALSA_IS_MAILBOX_NODE(mbnode));
+
+    mbnode->subscribed = !!subscribed;
+}
+
+/*
+ * Getters
+ */
+
+LibBalsaMailbox *
+balsa_mailbox_node_get_mailbox(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), NULL);
+
+    return mbnode->mailbox;
+}
+
+const gchar *
+balsa_mailbox_node_get_name(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), NULL);
+
+    return mbnode->name;
+}
+
+const gchar *
+balsa_mailbox_node_get_config_prefix(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), NULL);
+
+    return mbnode->config_prefix;
+}
+
+time_t
+balsa_mailbox_node_get_last_use_time(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), 0);
+
+    return mbnode->last_use;
+}
+
+LibBalsaServer *
+balsa_mailbox_node_get_server(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), NULL);
+
+    return mbnode->server;
+}
+
+BalsaMailboxNodeStyle
+balsa_mailbox_node_get_style(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), 0);
+
+    return mbnode->style;
+}
+
+guint
+balsa_mailbox_node_get_subscribed(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), 0);
+
+    return mbnode->subscribed;
+}
+
+guint
+balsa_mailbox_node_get_list_inbox(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), 0);
+
+    return mbnode->list_inbox;
+}
+
+gint
+balsa_mailbox_node_get_delim(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), 0);
+
+    return mbnode->delim;
+}
+
+const gchar *
+balsa_mailbox_node_get_dir(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), NULL);
+
+    return mbnode->dir;
+}
+
+BalsaMailboxNode *
+balsa_mailbox_node_get_parent(BalsaMailboxNode * mbnode)
+{
+    g_return_val_if_fail(BALSA_IS_MAILBOX_NODE(mbnode), NULL);
+
+    return mbnode->parent;
 }
