@@ -1,6 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- * Copyright (C) 1997-2016 Stuart Parmenter and others,
+ * Copyright (C) 1997-2019 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -14,7 +14,7 @@
  * GNU General Public License for more details.
  *  
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
  */
 
 #if defined(HAVE_CONFIG_H) && HAVE_CONFIG_H
@@ -26,12 +26,14 @@
 #include "balsa-app.h"
 #include "print.h"
 #include "misc.h"
+#include "html.h"
 #include "balsa-message.h"
 #include "quote-color.h"
 #include <glib/gi18n.h>
 #include "balsa-print-object.h"
 #include "balsa-print-object-decor.h"
 #include "balsa-print-object-header.h"
+#include "libbalsa-gpgme.h"
 
 #if HAVE__NL_MEASUREMENT_MEASUREMENT
 #include <langinfo.h>
@@ -47,9 +49,15 @@ typedef struct {
     GtkWidget *margin_bottom;
     GtkWidget *margin_left;
     GtkWidget *margin_right;
+#ifdef HAVE_HTML_WIDGET
+    GtkWidget *html_print;
+    GtkWidget *html_load_imgs;
+    gboolean prefer_text;
+    gboolean load_images;
+    BalsaPrintSetup *setup;
+#endif
 } BalsaPrintPrefs;
 
-
 typedef struct {
     /* related message */
     LibBalsaMessage *message;
@@ -91,7 +99,7 @@ print_header_footer(GtkPrintContext * context, cairo_t * cairo_ctx,
     g_free(pagebuf);
     cairo_move_to(cairo_ctx, pdata->setup.c_x0, pdata->c_header_y);
     pango_cairo_show_layout(cairo_ctx, layout);
-    g_object_unref(G_OBJECT(layout));
+    g_object_unref(layout);
 
     /* footer (if available) */
     if (pdata->footer) {
@@ -106,11 +114,91 @@ print_header_footer(GtkPrintContext * context, cairo_t * cairo_ctx,
 	pango_layout_set_text(layout, pdata->footer, -1);
 	cairo_move_to(cairo_ctx, pdata->setup.c_x0, pdata->c_footer_y);
 	pango_cairo_show_layout(cairo_ctx, layout);
-	g_object_unref(G_OBJECT(layout));
+	g_object_unref(layout);
     }
 }
 
 
+/*
+ * Scan the parts of a multipart/alternative as to find the proper one for printing.  According to RFC 2046, Sect. 5.1.4, the first
+ * /should/ be the "plain" one, and the following the "fancier" ones.  This is not guaranteed, though.
+ * 
+ * Furthermore, we may have cases like
+ * +- multipart/alternative
+ *      +- text/plain
+ *      +- multipart/mixed
+ *           +- text/html
+ *           +- ...
+ */
+static LibBalsaMessageBody *
+find_alt_part(LibBalsaMessageBody *parts,
+			  gboolean			   print_alt_html)
+{
+	LibBalsaMessageBody *part;
+	LibBalsaMessageBody *use_part = parts;		/* fallback: print 1st alternative part */
+
+	/* scan the parts */
+	for (part = parts; part != NULL; part = part->next) {
+		gchar *mime_type;
+		
+		mime_type = libbalsa_message_body_get_mime_type(part);
+		if ((strncmp(mime_type, "multipart/", 10U) == 0) && (part->parts != NULL)) {
+			/* consider the first child of a multipart */
+			g_free(mime_type);
+			mime_type = libbalsa_message_body_get_mime_type(part->parts);
+		}
+
+		if ((strcmp(mime_type, "text/calendar") == 0) ||
+			((strcmp(mime_type, "text/plain") == 0) && !print_alt_html) ||
+			((strcmp(mime_type, "text/html") == 0) && print_alt_html)) {
+			use_part = part;
+		}
+		g_free(mime_type);
+	}
+	
+	return use_part;
+}
+
+
+static inline GList *
+begin_crypto_frame(GList 		   *bpo_list,
+				   BalsaPrintSetup *psetup,
+				   gboolean 		is_signed,
+				   gboolean 		is_encrypted)
+{
+	if (is_signed) {
+		if (is_encrypted) {
+			bpo_list = balsa_print_object_frame_begin(bpo_list, _("Signed and encrypted"), psetup);
+		} else {
+			bpo_list = balsa_print_object_frame_begin(bpo_list, _("Signed"), psetup);
+		}
+	} else if (is_encrypted) {
+		bpo_list = balsa_print_object_frame_begin(bpo_list, _("Encrypted"), psetup);
+	} else {
+		/* no crypto, nothing to do... */
+	}
+	return bpo_list;
+}
+
+
+static GList *
+print_single_part(GList 			  *bpo_list,
+				  GtkPrintContext     *context,
+				  BalsaPrintSetup     *psetup,
+				  LibBalsaMessageBody *body,
+				  gboolean 			   no_first_sep,
+				  gboolean			   add_signature)
+{
+	if (!no_first_sep) {
+		bpo_list = balsa_print_object_separator(bpo_list, psetup);
+	}
+	if (add_signature) {
+		bpo_list = begin_crypto_frame(bpo_list, psetup, TRUE, body->was_encrypted);
+	}
+	return balsa_print_objects_append_from_body(bpo_list, context, body, psetup);
+}
+
+
 /*
  * scan the body list and prepare print data according to the content type
  */
@@ -118,91 +206,65 @@ static GList *
 scan_body(GList *bpo_list, GtkPrintContext * context, BalsaPrintSetup * psetup,
 	  LibBalsaMessageBody * body, gboolean no_first_sep)
 {
-#ifdef HAVE_GPGME
-    gboolean add_signature;
+    gboolean add_signature = FALSE;
     gboolean have_crypto_frame;
-#endif				/* HAVE_GPGME */
-
+	gboolean is_mp_signed;
     while (body) {
 	gchar *conttype;
 
 	conttype = libbalsa_message_body_get_mime_type(body);
-#ifdef HAVE_GPGME
 	add_signature = body->sig_info &&
-	    g_ascii_strcasecmp(conttype, "application/pgp-signature") &&
-	    g_ascii_strcasecmp(conttype, "application/pkcs7-signature") &&
-	    g_ascii_strcasecmp(conttype, "application/x-pkcs7-signature");
-	if (!g_ascii_strcasecmp("multipart/signed", conttype) &&
-	    body->parts && body->parts->next
-	    && body->parts->next->sig_info) {
+	    strcmp(conttype, "application/pgp-signature") != 0 &&
+	    strcmp(conttype, "application/pkcs7-signature") != 0 &&
+	    strcmp(conttype, "application/x-pkcs7-signature") != 0;
+
+	is_mp_signed = libbalsa_message_body_multipart_signed(body);
+	if (is_mp_signed || (!add_signature && body->was_encrypted)) {
 	    have_crypto_frame = TRUE;
 	    bpo_list = balsa_print_object_separator(bpo_list, psetup);
 	    no_first_sep = TRUE;
-	    if (body->was_encrypted)
-		bpo_list = balsa_print_object_frame_begin(bpo_list,
-							  _("Signed and encrypted matter"),
-							  psetup);
-	    else
-		bpo_list = balsa_print_object_frame_begin(bpo_list,
-							  _("Signed matter"),
-							  psetup);
-	} else if (!add_signature && body->was_encrypted) {
-	    have_crypto_frame = TRUE;
-	    bpo_list = balsa_print_object_separator(bpo_list, psetup);
-	    no_first_sep = TRUE;
-	    bpo_list = balsa_print_object_frame_begin(bpo_list,
-						      _("Encrypted matter"),
-						      psetup);
-	} else
+	    bpo_list = begin_crypto_frame(bpo_list, psetup, is_mp_signed, body->was_encrypted);
+	} else {
 	    have_crypto_frame = FALSE;
-#endif				/* HAVE_GPGME */
+	}
 
 	if (g_ascii_strncasecmp(conttype, "multipart/", 10)) {
-	    if (no_first_sep)
+		bpo_list = print_single_part(bpo_list, context, psetup, body, no_first_sep, add_signature);
 		no_first_sep = FALSE;
-	    else
-		bpo_list = balsa_print_object_separator(bpo_list, psetup);
-#ifdef HAVE_GPGME
-	    if (add_signature) {
-		if (body->was_encrypted)
-		    bpo_list = balsa_print_object_frame_begin(bpo_list,
-							      _("Signed and encrypted matter"),
-							      psetup);
-		else
-		    bpo_list = balsa_print_object_frame_begin(bpo_list,
-							      _("Signed matter"),
-							      psetup);
-	    }
-#endif				/* HAVE_GPGME */
-	    bpo_list = balsa_print_objects_append_from_body(bpo_list, context,
-							    body, psetup);
+	} else if (add_signature && !is_mp_signed) {
+		if (!no_first_sep) {
+			bpo_list = balsa_print_object_separator(bpo_list, psetup);
+			no_first_sep = TRUE;
+		}
+		bpo_list = begin_crypto_frame(bpo_list, psetup, TRUE, body->was_encrypted);
 	}
 
 	if (body->parts) {
-	    bpo_list = scan_body(bpo_list, context, psetup, body->parts, no_first_sep);
-	    no_first_sep = FALSE;
+		if (g_ascii_strcasecmp(conttype, "multipart/alternative") == 0) {
+			LibBalsaMessageBody *print_part;
+
+			print_part = find_alt_part(body->parts, psetup->print_alt_html);
+			bpo_list = print_single_part(bpo_list, context, psetup, print_part, no_first_sep, add_signature);
+		} else if (g_ascii_strcasecmp(conttype, "multipart/related") == 0) {
+			/* catch the case of a RFC 2387 multipart/related, typically a text/html with images which are enclosed in the
+			 * "related" container */
+			bpo_list = print_single_part(bpo_list, context, psetup, libbalsa_message_body_mp_related_root(body), no_first_sep,
+				add_signature);
+		} else {
+			bpo_list = scan_body(bpo_list, context, psetup, body->parts, no_first_sep);
+		}
+		no_first_sep = FALSE;
 	}
 
 	/* end the frame for an embedded message or encrypted stuff */
-	if (!g_ascii_strcasecmp(conttype, "message/rfc822")
-#ifdef HAVE_GPGME
-	    || have_crypto_frame
-#endif
-	    )
+	if (!g_ascii_strcasecmp(conttype, "message/rfc822") || have_crypto_frame)
 	    bpo_list = balsa_print_object_frame_end(bpo_list, psetup);
 
-#ifdef HAVE_GPGME
 	if (add_signature) {
-	    gchar *header =
-		g_strdup_printf(_("This is an inline %s signed %s message part:"),
-				body->sig_info->protocol == GPGME_PROTOCOL_OpenPGP ?
-				_("OpenPGP") : _("S/MIME"), conttype);
 	    bpo_list = balsa_print_object_separator(bpo_list, psetup);
-	    bpo_list = balsa_print_object_header_crypto(bpo_list, context, body, header, psetup);
-	    g_free(header);
+	    bpo_list = balsa_print_object_header_crypto(bpo_list, context, body, psetup);
 	    bpo_list = balsa_print_object_frame_end(bpo_list, psetup);
 	}
-#endif				/* HAVE_GPGME */
 	g_free(conttype);
 
 	body = body->next;
@@ -211,7 +273,6 @@ scan_body(GList *bpo_list, GtkPrintContext * context, BalsaPrintSetup * psetup,
     return bpo_list;
 }
 
-
 static void
 begin_print(GtkPrintOperation * operation, GtkPrintContext * context,
 	    BalsaPrintData * pdata)
@@ -223,6 +284,7 @@ begin_print(GtkPrintOperation * operation, GtkPrintContext * context,
     gchar *subject;
     gchar *date;
     GString *footer_string;
+    InternetAddressList *from_list;
 
     /* initialise the context */
     page_setup = gtk_print_context_get_page_setup(context);
@@ -277,16 +339,19 @@ begin_print(GtkPrintOperation * operation, GtkPrintContext * context,
 	footer_string = NULL;
 
     date = libbalsa_message_date_to_utf8(pdata->message, balsa_app.date_string);
-    if (footer_string) {
-	footer_string = g_string_append(footer_string, " \342\200\224 ");
-	footer_string = g_string_append(footer_string, date);
-    } else
-	footer_string = g_string_new(date);
-    g_free(date);
+    if (date != NULL) {
+    	if (footer_string != NULL) {
+    		footer_string = g_string_append(footer_string, " \342\200\224 ");
+    		footer_string = g_string_append(footer_string, date);
+    	} else {
+    		footer_string = g_string_new(date);
+    	}
+    	g_free(date);
+    }
 
-    if (pdata->message->headers->from) {
-	gchar *from =
-	    internet_address_list_to_string(pdata->message->headers->from, FALSE);
+    from_list = libbalsa_message_get_headers(pdata->message)->from;
+    if (from_list != NULL) {
+	gchar *from = internet_address_list_to_string(from_list, NULL, FALSE);
 
 	libbalsa_utf8_sanitize(&from, balsa_app.convert_unknown_8bit,
 			       NULL);
@@ -320,7 +385,7 @@ begin_print(GtkPrintOperation * operation, GtkPrintContext * context,
 	/* remember in the context */
 	pdata->footer = g_string_free(footer_string, FALSE);
     }
-    g_object_unref(G_OBJECT(layout));
+    g_object_unref(layout);
 
     /* add the message headers */
     pdata->setup.c_y_pos = 0.0;	/* to simplify calculating the layout... */
@@ -332,7 +397,7 @@ begin_print(GtkPrintOperation * operation, GtkPrintContext * context,
     /* add the mime bodies */
     pdata->print_parts = 
 	scan_body(pdata->print_parts, context, &pdata->setup,
-		  pdata->message->body_list, FALSE);
+		  libbalsa_message_get_body_list(pdata->message), FALSE);
 
     /* done */
     gtk_print_operation_set_n_pages(operation, pdata->setup.page_count);
@@ -369,7 +434,7 @@ draw_page(GtkPrintOperation * operation, GtkPrintContext * context,
     while (p) {
 	BalsaPrintObject *po = BALSA_PRINT_OBJECT(p->data);
 
-	if (po->on_page == page_nr)
+	if (balsa_print_object_get_page(po) == page_nr)
 	    balsa_print_object_draw(po, context, cairo_ctx);
 
 	p = g_list_next(p);
@@ -402,7 +467,7 @@ get_default_user_units(void)
     if (strcmp(e, "default:inch")==0)
 	return GTK_UNIT_INCH;
     else if (strcmp(e, "default:mm"))
-	g_warning("Whoever translated default:mm did so wrongly.\n");
+	g_warning("Whoever translated default:mm did so wrongly.");
     return GTK_UNIT_MM;
 }
 
@@ -414,15 +479,15 @@ add_font_button(const gchar * text, const gchar * font, GtkGrid * grid,
     GtkWidget *font_button;
 
     label = gtk_label_new_with_mnemonic(text);
-    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
     gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
     gtk_widget_set_halign(label, GTK_ALIGN_START);
-    gtk_grid_attach(grid, label, 0, row, 1, 1);
+    gtk_label_set_xalign(GTK_LABEL(label), 0.0F);
+    gtk_grid_attach(grid, label, 1, row, 1, 1);
 
     font_button = gtk_font_button_new_with_font(font);
     gtk_label_set_mnemonic_widget(GTK_LABEL(label), font_button);
     gtk_widget_set_hexpand(font_button, TRUE);
-    gtk_grid_attach(grid, font_button, 1, row, 1, 1);
+    gtk_grid_attach(grid, font_button, 2, row, 1, 1);
 
     return font_button;
 }
@@ -437,10 +502,10 @@ add_margin_spinbtn(const gchar * text, gdouble min, gdouble max, gdouble dflt,
     const gchar *unit;
 
     label = gtk_label_new_with_mnemonic(text);
-    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
     gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
     gtk_widget_set_halign(label, GTK_ALIGN_START);
-    gtk_grid_attach(grid, label, 0, row, 1, 1);
+    gtk_label_set_xalign(GTK_LABEL(label), 0.0F);
+    gtk_grid_attach(grid, label, 1, row, 1, 1);
 
     if (get_default_user_units() == GTK_UNIT_INCH) {
 	unit = _("inch");
@@ -459,13 +524,11 @@ add_margin_spinbtn(const gchar * text, gdouble min, gdouble max, gdouble dflt,
     }
     gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(spinbtn), TRUE);
     gtk_label_set_mnemonic_widget(GTK_LABEL(label), spinbtn);
-    gtk_grid_attach(grid, spinbtn, 1, row, 1, 1);
+    gtk_grid_attach(grid, spinbtn, 2, row, 1, 1);
 
     label = gtk_label_new(unit);
-    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
-    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
     gtk_widget_set_halign(label, GTK_ALIGN_START);
-    gtk_grid_attach(grid, label, 2, row, 1, 1);
+    gtk_grid_attach(grid, label, 3, row, 1, 1);
 
     return spinbtn;
 }
@@ -481,116 +544,87 @@ check_margins(GtkAdjustment * adjustment, GtkAdjustment * other)
                                  gtk_adjustment_get_value(other));
 }
 
+static GtkWidget *
+create_options_group(const gchar *label_str, GtkWidget *parent_grid, gint parent_col, gint parent_row, gint parent_width)
+{
+	GtkWidget *group;
+	GtkWidget *label;
+    GtkWidget *grid;
+    gchar *markup;
+
+    group = gtk_box_new(GTK_ORIENTATION_VERTICAL, 12);
+    gtk_grid_attach(GTK_GRID(parent_grid), group, parent_col, parent_row, parent_width, 1);
+
+    markup = g_strdup_printf("<b>%s</b>", label_str);
+    label = libbalsa_create_wrap_label(markup, TRUE);
+    g_free(markup);
+    gtk_box_pack_start(GTK_BOX(group), label, FALSE, FALSE, 0);
+
+	grid = gtk_grid_new();
+    gtk_box_pack_start(GTK_BOX(group), grid, FALSE, FALSE, 0);
+	gtk_grid_set_column_spacing(GTK_GRID(grid), 0);
+	gtk_grid_set_row_spacing(GTK_GRID(grid), 6);
+	gtk_grid_attach(GTK_GRID(grid), gtk_label_new("    "), 0, 0, 1, 1);
+	return grid;
+}
+
 static GtkWidget *
 message_prefs_widget(GtkPrintOperation * operation,
 		     BalsaPrintPrefs * print_prefs)
 {
     GtkWidget *page;
-    GtkWidget *group;
-    GtkWidget *label;
-    GtkWidget *hbox;
-    GtkWidget *vbox;
     GtkWidget *grid;
+#ifdef HAVE_HTML_WIDGET
+    GtkWidget *dummy;
+#endif
     GtkPageSetup *pg_setup;
-    gchar *markup;
 
     gtk_print_operation_set_custom_tab_label(operation, _("Message"));
 
-    page = gtk_box_new(GTK_ORIENTATION_VERTICAL, 18);
+    page = gtk_grid_new();
+    gtk_grid_set_column_spacing(GTK_GRID(page), 18);
+    gtk_grid_set_row_spacing(GTK_GRID(page), 18);
     gtk_container_set_border_width(GTK_CONTAINER(page), 12);
 
-    group = gtk_box_new(GTK_ORIENTATION_VERTICAL, 12);
-    gtk_box_pack_start(GTK_BOX(page), group, FALSE, TRUE, 0);
+    /* fonts */
+    grid = create_options_group(_("Fonts"), page, 0, 0, 3);
 
-    label = gtk_label_new(NULL);
-    markup = g_strdup_printf("<b>%s</b>", _("Fonts"));
-    gtk_label_set_markup(GTK_LABEL(label), markup);
-    g_free(markup);
-    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-    gtk_widget_set_halign(label, GTK_ALIGN_START);
-    gtk_box_pack_start(GTK_BOX(group), label, FALSE, FALSE, 0);
+    print_prefs->header_font = add_font_button(_("_Header Font:"), balsa_app.print_header_font, GTK_GRID(grid), 0);
+    print_prefs->body_font = add_font_button(_("B_ody Font:"), balsa_app.print_body_font, GTK_GRID(grid), 1);
+    print_prefs->footer_font = add_font_button(_("_Footer Font:"), balsa_app.print_footer_font, GTK_GRID(grid), 2);
 
-    hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
-    gtk_box_pack_start(GTK_BOX(group), hbox, TRUE, TRUE, 0);
-    gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("    "),
-		       FALSE, FALSE, 0);
-    vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 6);
-    gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
-
-    grid = gtk_grid_new();
-    gtk_grid_set_row_spacing(GTK_GRID(grid), 6);
-    gtk_grid_set_column_spacing(GTK_GRID(grid), 6);
-
-    gtk_box_pack_start(GTK_BOX(vbox), grid, FALSE, TRUE, 0);
-
-    print_prefs->header_font =
-	add_font_button(_("_Header Font:"), balsa_app.print_header_font,
-			GTK_GRID(grid), 0);
-    print_prefs->body_font =
-	add_font_button(_("B_ody Font:"), balsa_app.print_body_font,
-			GTK_GRID(grid), 1);
-    print_prefs->footer_font =
-	add_font_button(_("_Footer Font:"), balsa_app.print_footer_font,
-			GTK_GRID(grid), 2);
-
-    group = gtk_box_new(GTK_ORIENTATION_VERTICAL, 12);
-    gtk_box_pack_start(GTK_BOX(page), group, FALSE, TRUE, 0);
-
-    label = gtk_label_new(NULL);
-    markup = g_strdup_printf("<b>%s</b>", _("Highlighting"));
-    gtk_label_set_markup(GTK_LABEL(label), markup);
-    g_free(markup);
-    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-    gtk_widget_set_halign(label, GTK_ALIGN_START);
-    gtk_box_pack_start(GTK_BOX(group), label, FALSE, FALSE, 0);
-
-    hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
-    gtk_box_pack_start(GTK_BOX(group), hbox, TRUE, TRUE, 0);
-    gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("    "),
-		       FALSE, FALSE, 0);
-    vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 6);
-    gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
+    /* syntax highlighting */
+    grid = create_options_group(_("Highlighting"), page, 0, 1, 1);
 
     print_prefs->highlight_cited =
 	gtk_check_button_new_with_mnemonic(_("Highlight _cited text"));
-    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON
-				 (print_prefs->highlight_cited),
-				 balsa_app.print_highlight_cited);
-    gtk_box_pack_start(GTK_BOX(vbox), print_prefs->highlight_cited, FALSE,
-		       TRUE, 0);
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(print_prefs->highlight_cited), balsa_app.print_highlight_cited);
+    gtk_grid_attach(GTK_GRID(grid), print_prefs->highlight_cited, 1, 0, 1, 1);
 
-    print_prefs->highlight_phrases =
-	gtk_check_button_new_with_mnemonic(_
-					   ("Highlight _structured phrases"));
-    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON
-				 (print_prefs->highlight_phrases),
-				 balsa_app.print_highlight_phrases);
-    gtk_box_pack_start(GTK_BOX(vbox), print_prefs->highlight_phrases,
-		       FALSE, TRUE, 0);
+    print_prefs->highlight_phrases = gtk_check_button_new_with_mnemonic(_("Highlight _structured phrases"));
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(print_prefs->highlight_phrases), balsa_app.print_highlight_phrases);
+    gtk_grid_attach(GTK_GRID(grid), print_prefs->highlight_phrases, 1, 1, 1, 1);
 
-    group = gtk_box_new(GTK_ORIENTATION_VERTICAL, 12);
-    gtk_box_pack_start(GTK_BOX(page), group, FALSE, TRUE, 0);
+#ifdef HAVE_HTML_WIDGET
+    /* treatment of HTML messages and parts */
+    grid = create_options_group(_("Highlighting"), page, 1, 1, 1);
 
-    label = gtk_label_new(NULL);
-    markup = g_strdup_printf("<b>%s</b>", _("Margins"));
-    gtk_label_set_markup(GTK_LABEL(label), markup);
-    g_free(markup);
-    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-    gtk_widget_set_halign(label, GTK_ALIGN_START);
-    gtk_box_pack_start(GTK_BOX(group), label, FALSE, FALSE, 0);
+    print_prefs->html_print = gtk_check_button_new_with_mnemonic(_("Prefer text/plain over HTML"));
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(print_prefs->html_print), print_prefs->prefer_text);
+    gtk_grid_attach(GTK_GRID(grid), print_prefs->html_print, 1, 0, 1, 1);
 
-    hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
-    gtk_box_pack_start(GTK_BOX(group), hbox, TRUE, TRUE, 0);
-    gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("    "),
-		       FALSE, FALSE, 0);
-    vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 6);
-    gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
+    print_prefs->html_load_imgs = gtk_check_button_new_with_mnemonic(_("Download images from remote servers (may be dangerous)"));
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(print_prefs->html_load_imgs), print_prefs->load_images);
+    gtk_grid_attach(GTK_GRID(grid), print_prefs->html_load_imgs, 1, 1, 1, 1);
 
-    grid = gtk_grid_new();
-    gtk_grid_set_row_spacing(GTK_GRID(grid), 6);
-    gtk_grid_set_column_spacing(GTK_GRID(grid), 6);
+    /* phantom alignment */
+    dummy = gtk_label_new(" ");
+    gtk_widget_set_hexpand(dummy, TRUE);
+    gtk_grid_attach(GTK_GRID(grid), dummy, 2, 1, 1, 1);
+#endif
 
-    gtk_box_pack_start(GTK_BOX(vbox), grid, FALSE, TRUE, 0);
+    /* margins */
+    grid = create_options_group(_("Margins"), page, 0, 2, 2);
 
     pg_setup = gtk_print_operation_get_default_page_setup(operation);
     print_prefs->margin_top =
@@ -605,10 +639,10 @@ message_prefs_widget(GtkPrintOperation * operation,
 			   gtk_page_setup_get_page_height(pg_setup, GTK_UNIT_POINTS), 
 			   balsa_app.margin_bottom,
 			   GTK_GRID(grid), 1);
-    g_signal_connect(G_OBJECT(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_top))),
+    g_signal_connect(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_top)),
 		     "value-changed", G_CALLBACK(check_margins),
 		     gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_bottom)));
-    g_signal_connect(G_OBJECT(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_bottom))),
+    g_signal_connect(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_bottom)),
 		     "value-changed", G_CALLBACK(check_margins),
 		     gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_top)));
     print_prefs->margin_left =
@@ -623,10 +657,10 @@ message_prefs_widget(GtkPrintOperation * operation,
 			   gtk_page_setup_get_page_width(pg_setup, GTK_UNIT_POINTS), 
 			   balsa_app.margin_right,
 			   GTK_GRID(grid), 3);
-    g_signal_connect(G_OBJECT(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_left))),
+    g_signal_connect(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_left)),
 		     "value-changed", G_CALLBACK(check_margins),
 		     gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_right)));
-    g_signal_connect(G_OBJECT(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_right))),
+    g_signal_connect(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_right)),
 		     "value-changed", G_CALLBACK(check_margins),
 		     gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(print_prefs->margin_left)));
 
@@ -671,6 +705,12 @@ message_prefs_apply(GtkPrintOperation * operation, GtkWidget * widget,
 	balsa_app.margin_left /= 25.4;
 	balsa_app.margin_right /= 25.4;
     }
+#ifdef HAVE_HTML_WIDGET
+    print_prefs->setup->print_alt_html =
+    	!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(print_prefs->html_print));
+    print_prefs->setup->html_load_images =
+    	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(print_prefs->html_load_imgs));
+#endif
 }
 
 
@@ -687,7 +727,7 @@ message_print_page_setup(GtkWindow * parent)
 					balsa_app.print_settings);
 
     if (balsa_app.page_setup)
-	g_object_unref(G_OBJECT(balsa_app.page_setup));
+	g_object_unref(balsa_app.page_setup);
 
     balsa_app.page_setup = new_page_setup;
 }
@@ -700,12 +740,15 @@ message_print(LibBalsaMessage * msg, GtkWindow * parent)
     GtkPrintOperationResult res;
     BalsaPrintData *print_data;
     BalsaPrintPrefs print_prefs;
+#ifdef HAVE_HTML_WIDGET
+    InternetAddressList *from;
+#endif
     GError *err = NULL;
 
     print = gtk_print_operation_new();
     g_assert(print != NULL);
 
-    g_object_ref(G_OBJECT(msg));
+    g_object_ref(msg);
 
     gtk_print_operation_set_n_pages(print, 1);
     gtk_print_operation_set_unit(print, GTK_UNIT_POINTS);
@@ -721,6 +764,12 @@ message_print(LibBalsaMessage * msg, GtkWindow * parent)
     /* create a print context */
     print_data = g_new0(BalsaPrintData, 1);
     print_data->message = msg;
+#ifdef HAVE_HTML_WIDGET
+    print_prefs.setup = &print_data->setup;
+    from = libbalsa_message_get_headers(msg)->from;
+    print_prefs.prefer_text = balsa_app.display_alt_plain && !libbalsa_html_get_prefer_html(from);
+    print_prefs.load_images = libbalsa_html_get_load_images(from);
+#endif
 
     g_signal_connect(print, "begin_print", G_CALLBACK(begin_print), print_data);
     g_signal_connect(print, "draw_page", G_CALLBACK(draw_page), print_data);
@@ -746,10 +795,9 @@ message_print(LibBalsaMessage * msg, GtkWindow * parent)
     /* clean up */
     if (err)
 	g_error_free(err);
-    g_list_foreach(print_data->print_parts, (GFunc) g_object_unref, NULL);
-    g_list_free(print_data->print_parts);
+    g_list_free_full(print_data->print_parts, g_object_unref);
     g_free(print_data->footer);
     g_free(print_data);
-    g_object_unref(G_OBJECT(print));
-    g_object_unref(G_OBJECT(msg));
+    g_object_unref(print);
+    g_object_unref(msg);
 }
